from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtBluetooth
from prettyqt.utils import bidict


md = QtBluetooth.QBluetoothUuid.CharacteristicType


CharacteristicTypeStr = Literal[
    "aerobic_heart_rate_lower",
    "aerobic_heart_rate_upper",
    "aerobic_threshold",
    "age",
    "anaerobic_heart_rate_lower",
    "anaerobic_heart_rate_upper",
    "anaerobic_threshold",
    "alert_category_id",
    "alert_category_id_bitmask",
    "alert_level",
    "alert_notification_control_point",
    "alert_status",
    "apparent_wind_direction",
    "apparent_wind_speed",
    "appearance",
    "barometric_pressure_trend",
    "battery_level",
    "blood_pressure_feature",
    "blood_pressure_measurement",
    "body_composition_feature",
    "body_composition_measurement",
    "body_sensor_location",
    "boot_keyboard_input_report",
    "boot_keyboard_output_report",
    "boot_mouse_input_report",
    "csc_feature",
    "csc_measurement",
    "current_time",
    "cycling_power_control_point",
    "cycling_power_feature",
    "cycling_power_measurement",
    "cycling_power_vector",
    "database_change_increment",
    "date_of_birth",
    "date_of_threshold_assessment",
    "date_time",
    "day_date_time",
    "day_of_week",
    "descriptor_value_changed",
    "device_name",
    "dew_point",
    "dst_offset",
    "elevation",
    "email_address",
    "exact_time_256",
    "fat_burn_heart_rate_lower_limit",
    "fat_burn_heart_rate_upper_limit",
    "firmware_revision_string",
    "first_name",
    "five_zone_heart_rate_limits",
    "gender",
    "glucose_feature",
    "glucose_measurement",
    "glucose_measurement_context",
    "gust_factor",
    "hardware_revision_string",
    "max_recommended_heart_rate",
    "heart_rate_control_point",
    "heart_rate_max",
    "heart_rate_measurement",
    "heat_index",
    "height",
    "hid_control_point",
    "hid_information",
    "hip_circum_reference",
    "humidity",
    "regulatory_certification_list",
    "intermediate_cuff_pressure",
    "intermediate_temperature",
    "irradiance",
    "language",
    "last_name",
    "ln_control_point",
    "ln_feature",
    "local_time_information",
    "location_and_speed",
    "magnetic_declination",
    "magnetic_flux_density_2d",
    "magnetic_magnetic_flux_density_3d",
    "manufacturer_name_string",
    "measurement_interval",
    "model_number_string",
    "navigation",
    "new_alert",
    "peripheral_preferred_connection_parameters",
    "peripheral_privacy_flag",
    "pn_pid",
    "pollen_concentration",
    "position_quality",
    "pressure",
    "protocol_mode",
    "rainfall",
    "reconnection_address",
    "record_access_control_point",
    "reference_time_information",
    "report",
    "report_map",
    "resting_heart_rate",
    "ringer_control_point",
    "ringer_setting",
    "rsc_feature",
    "rsc_measurement",
    "sc_control_point",
    "scan_interval_window",
    "scan_refresh",
    "sensor_location",
    "serial_number_string",
    "service_changed",
    "software_revision_string",
    "sport_type_for_aerobic_anaerobic_thresholds",
    "supported_new_alert_category",
    "supported_unread_alert_category",
    "system_id",
    "temperature",
    "temperature_measurement",
    "temperature_type",
    "three_zone_heart_rate_limits",
    "time_accuracy",
    "time_source",
    "time_update_control_point",
    "time_update_state",
    "time_with_dst",
    "time_zone",
    "true_wind_direction",
    "true_wind_speed",
    "two_zone_hear_rate_limits",
    "tx_power_level",
    "unread_alert_status",
    "user_control_point",
    "user_index",
    "uv_index",
    "vo_2_max",
    "waist_circumference",
    "weight",
    "weight_measurement",
    "weight_scale_feature",
    "wind_chill",
]

CHARACTERISTIC_TYPES: bidict[CharacteristicTypeStr, md] = bidict(
    aerobic_heart_rate_lower=md.AerobicHeartRateLowerLimit,
    aerobic_heart_rate_upper=md.AerobicHeartRateUpperLimit,
    aerobic_threshold=md.AerobicThreshold,
    age=md.Age,
    anaerobic_heart_rate_lower=md.AnaerobicHeartRateLowerLimit,
    anaerobic_heart_rate_upper=md.AnaerobicHeartRateUpperLimit,
    anaerobic_threshold=md.AnaerobicThreshold,
    alert_category_id=md.AlertCategoryID,
    alert_category_id_bitmask=md.AlertCategoryIDBitMask,
    alert_level=md.AlertLevel,
    alert_notification_control_point=md.AlertNotificationControlPoint,
    alert_status=md.AlertStatus,
    apparent_wind_direction=md.ApparentWindDirection,
    apparent_wind_speed=md.ApparentWindSpeed,
    appearance=md.Appearance,
    barometric_pressure_trend=md.BarometricPressureTrend,
    battery_level=md.BatteryLevel,
    blood_pressure_feature=md.BloodPressureFeature,
    blood_pressure_measurement=md.BloodPressureMeasurement,
    body_composition_feature=md.BodyCompositionFeature,
    body_composition_measurement=md.BodyCompositionMeasurement,
    body_sensor_location=md.BodySensorLocation,
    boot_keyboard_input_report=md.BootKeyboardInputReport,
    boot_keyboard_output_report=md.BootKeyboardOutputReport,
    boot_mouse_input_report=md.BootMouseInputReport,
    csc_feature=md.CSCFeature,
    csc_measurement=md.CSCMeasurement,
    current_time=md.CurrentTime,
    cycling_power_control_point=md.CyclingPowerControlPoint,
    cycling_power_feature=md.CyclingPowerFeature,
    cycling_power_measurement=md.CyclingPowerMeasurement,
    cycling_power_vector=md.CyclingPowerVector,
    database_change_increment=md.DatabaseChangeIncrement,
    date_of_birth=md.DateOfBirth,
    date_of_threshold_assessment=md.DateOfThresholdAssessment,
    date_time=md.DateTime,
    day_date_time=md.DayDateTime,
    day_of_week=md.DayOfWeek,
    descriptor_value_changed=md.DescriptorValueChanged,
    device_name=md.DeviceName,
    dew_point=md.DewPoint,
    dst_offset=md.DSTOffset,
    elevation=md.Elevation,
    email_address=md.EmailAddress,
    exact_time_256=md.ExactTime256,
    fat_burn_heart_rate_lower_limit=md.FatBurnHeartRateLowerLimit,
    fat_burn_heart_rate_upper_limit=md.FatBurnHeartRateUpperLimit,
    firmware_revision_string=md.FirmwareRevisionString,
    first_name=md.FirstName,
    five_zone_heart_rate_limits=md.FiveZoneHeartRateLimits,
    gender=md.Gender,
    glucose_feature=md.GlucoseFeature,
    glucose_measurement=md.GlucoseMeasurement,
    glucose_measurement_context=md.GlucoseMeasurementContext,
    gust_factor=md.GustFactor,
    hardware_revision_string=md.HardwareRevisionString,
    max_recommended_heart_rate=md.MaximumRecommendedHeartRate,
    heart_rate_control_point=md.HeartRateControlPoint,
    heart_rate_max=md.HeartRateMax,
    heart_rate_measurement=md.HeartRateMeasurement,
    heat_index=md.HeatIndex,
    height=md.Height,
    hid_control_point=md.HIDControlPoint,
    hid_information=md.HIDInformation,
    hip_circum_reference=md.HipCircumference,
    humidity=md.Humidity,
    regulatory_certification_list=md.IEEE1107320601RegulatoryCertificationDataList,
    intermediate_cuff_pressure=md.IntermediateCuffPressure,
    intermediate_temperature=md.IntermediateTemperature,
    irradiance=md.Irradiance,
    language=md.Language,
    last_name=md.LastName,
    ln_control_point=md.LNControlPoint,
    ln_feature=md.LNFeature,
    local_time_information=md.LocalTimeInformation,
    location_and_speed=md.LocationAndSpeed,
    magnetic_declination=md.MagneticDeclination,
    magnetic_flux_density_2d=md.MagneticFluxDensity2D,
    magnetic_magnetic_flux_density_3d=md.MagneticFluxDensity3D,
    manufacturer_name_string=md.ManufacturerNameString,
    measurement_interval=md.MeasurementInterval,
    model_number_string=md.ModelNumberString,
    navigation=md.Navigation,
    new_alert=md.NewAlert,
    peripheral_preferred_connection_parameters=md.PeripheralPreferredConnectionParameters,
    peripheral_privacy_flag=md.PeripheralPrivacyFlag,
    pn_pid=md.PnPID,
    pollen_concentration=md.PollenConcentration,
    position_quality=md.PositionQuality,
    pressure=md.Pressure,
    protocol_mode=md.ProtocolMode,
    rainfall=md.Rainfall,
    reconnection_address=md.ReconnectionAddress,
    record_access_control_point=md.RecordAccessControlPoint,
    reference_time_information=md.ReferenceTimeInformation,
    report=md.Report,
    report_map=md.ReportMap,
    resting_heart_rate=md.RestingHeartRate,
    ringer_control_point=md.RingerControlPoint,
    ringer_setting=md.RingerSetting,
    rsc_feature=md.RSCFeature,
    rsc_measurement=md.RSCMeasurement,
    sc_control_point=md.SCControlPoint,
    scan_interval_window=md.ScanIntervalWindow,
    scan_refresh=md.ScanRefresh,
    sensor_location=md.SensorLocation,
    serial_number_string=md.SerialNumberString,
    service_changed=md.ServiceChanged,
    software_revision_string=md.SoftwareRevisionString,
    sport_type_for_aerobic_anaerobic_thresholds=md.SportTypeForAerobicAnaerobicThresholds,
    supported_new_alert_category=md.SupportedNewAlertCategory,
    supported_unread_alert_category=md.SupportedUnreadAlertCategory,
    system_id=md.SystemID,
    temperature=md.Temperature,
    temperature_measurement=md.TemperatureMeasurement,
    temperature_type=md.TemperatureType,
    three_zone_heart_rate_limits=md.ThreeZoneHeartRateLimits,
    time_accuracy=md.TimeAccuracy,
    time_source=md.TimeSource,
    time_update_control_point=md.TimeUpdateControlPoint,
    time_update_state=md.TimeUpdateState,
    time_with_dst=md.TimeWithDST,
    time_zone=md.TimeZone,
    true_wind_direction=md.TrueWindDirection,
    true_wind_speed=md.TrueWindSpeed,
    two_zone_hear_rate_limits=md.TwoZoneHeartRateLimits,
    tx_power_level=md.TxPowerLevel,
    unread_alert_status=md.UnreadAlertStatus,
    user_control_point=md.UserControlPoint,
    user_index=md.UserIndex,
    uv_index=md.UVIndex,
    vo_2_max=md.VO2Max,
    waist_circumference=md.WaistCircumference,
    weight=md.Weight,
    weight_measurement=md.WeightMeasurement,
    weight_scale_feature=md.WeightScaleFeature,
    wind_chill=md.WindChill,
)

dt = QtBluetooth.QBluetoothUuid.DescriptorType

DescriptorTypeStr = Literal[
    "characteristic_extended_properties",
    "characteristic_user_description",
    "client_characteristic_configuration",
    "server_characteristic_configuration",
    "characteristic_presentation_format",
    "characteristic_aggregate_format",
    "valid_range",
    "external_report_reference",
    "report_reference",
    "environmental_sensing_configuration",
    "environmental_sensing_measurement",
    "environmental_sensing_trigger_setting",
    "unknown_descriptor_type",
]

DESCRIPTOR_TYPE: bidict[DescriptorTypeStr, dt] = bidict(
    characteristic_extended_properties=dt.CharacteristicExtendedProperties,
    characteristic_user_description=dt.CharacteristicUserDescription,
    client_characteristic_configuration=dt.ClientCharacteristicConfiguration,
    server_characteristic_configuration=dt.ServerCharacteristicConfiguration,
    characteristic_presentation_format=dt.CharacteristicPresentationFormat,
    characteristic_aggregate_format=dt.CharacteristicAggregateFormat,
    valid_range=dt.ValidRange,
    external_report_reference=dt.ExternalReportReference,
    report_reference=dt.ReportReference,
    environmental_sensing_configuration=dt.EnvironmentalSensingConfiguration,
    environmental_sensing_measurement=dt.EnvironmentalSensingMeasurement,
    environmental_sensing_trigger_setting=dt.EnvironmentalSensingTriggerSetting,
    unknown_descriptor_type=dt.UnknownDescriptorType,
)

pu = QtBluetooth.QBluetoothUuid.ProtocolUuid

ProtocolUuidstr = Literal[
    "sdp",
    "udp",
    "rfcomm",
    "tcp",
    "tcs_bin",
    "tcs_at",
    "att",
    "obex",
    "ip",
    "ftp",
    "http",
    "wsp",
    "bnep",
    "upnp",
    "hidp",
    "hardcopy_control_channel",
    "hardcopy_data_channel",
    "hardcopy_notification",
    "avctp",
    "avdtp",
    "cmtp",
    "udi_c_plain",
    "mcap_control_channel",
    "mcap_data_channel",
    "l2_cap",
]

PROTOCOL_UUID: bidict[ProtocolUuidstr, pu] = bidict(
    sdp=pu.Sdp,
    udp=pu.Udp,
    rfcomm=pu.Rfcomm,
    tcp=pu.Tcp,
    tcs_bin=pu.TcsBin,
    tcs_at=pu.TcsAt,
    att=pu.Att,
    obex=pu.Obex,
    ip=pu.Ip,
    ftp=pu.Ftp,
    http=pu.Http,
    wsp=pu.Wsp,
    bnep=pu.Bnep,
    upnp=pu.Upnp,
    hidp=pu.Hidp,
    hardcopy_control_channel=pu.HardcopyControlChannel,
    hardcopy_data_channel=pu.HardcopyDataChannel,
    hardcopy_notification=pu.HardcopyNotification,
    avctp=pu.Avctp,
    avdtp=pu.Avdtp,
    cmtp=pu.Cmtp,
    udi_c_plain=pu.UdiCPlain,
    mcap_control_channel=pu.McapControlChannel,
    mcap_data_channel=pu.McapDataChannel,
    l2_cap=pu.L2cap,
)

su = QtBluetooth.QBluetoothUuid.ServiceClassUuid

ServiceClassUuidStr = Literal[
    "service_discovery_server",
    "browser_group_descriptor",
    "public_browse_group",
    "serial_port",
    "lan_access_using_ppp",
    "dial_up_networking",
    "ir_mc_sync",
    "obex_object_push",
    "obex_file_transfer",
    "ir_mc_sync_command",
    "headset",
    "audio_source",
    "audio_sink",
    "av_remove_control_target",
    "advanced_audio_distribution",
    "av_remote_control",
    "av_remote_control_controller",
    "headset_ag",
    "panu",
    "nap",
    "gn",
    "direct_printing",
    "reference_printing",
    "basic_image",
    "imaging_responder",
    "imaging_automatic_archive",
    "imaging_reference_objects",
    "hands_free",
    "hands_free_audio_gateway",
    "direct_printing_reference_objects_service",
    "reflected_ui",
    "basic_printing",
    "printing_status",
    "human_interface_device_service",
    "hardcopy_cable_replacement",
    "hcr_print",
    "hcr_scan",
    "sim_access",
    "phonebook_access_pce",
    "phonebook_access_pse",
    "phonebook_access",
    "headset_hs",
    "message_access_server",
    "message_notification_server",
    "message_access_profile",
    "gnss",
    "gnss_server",
    "display_3d",
    "glasses_3d",
    "synchronization_3d",
    "mps_profile",
    "mps_service",
    "pnp_information",
    "generic_networking",
    "generic_file_transfer",
    "generic_audio",
    "generic_telephony",
    "video_source",
    "video_sink",
    "video_distribution",
    "hdp",
    "hdp_source",
    "hdp_sink",
    "generic_access",
    "generic_attribute",
    "immediate_alert",
    "link_loss",
    "tx_power",
    "current_time_service",
    "reference_time_update_service",
    "next_dst_change_service",
    "glucose",
    "health_thermometer",
    "device_information",
    "heart_rate",
    "phone_alert_status_service",
    "battery_service",
    "blood_pressure",
    "alert_notification_service",
    "human_interface_device",
    "scan_parameters",
    "running_speed_and_cadence",
    "cycling_speed_and_cadence",
    "cycling_power",
    "location_and_navigation",
    "environmental_sensing",
    "body_composition",
    "user_data",
    "weight_scale",
    "bond_management",
    "continuous_glucose_monitoring",
]

SERVICE_CLASS_UUID: bidict[ServiceClassUuidStr, su] = bidict(
    service_discovery_server=su.ServiceDiscoveryServer,
    browser_group_descriptor=su.BrowseGroupDescriptor,
    public_browse_group=su.PublicBrowseGroup,
    serial_port=su.SerialPort,
    lan_access_using_ppp=su.LANAccessUsingPPP,
    dial_up_networking=su.DialupNetworking,
    ir_mc_sync=su.IrMCSync,
    obex_object_push=su.ObexObjectPush,
    obex_file_transfer=su.OBEXFileTransfer,
    ir_mc_sync_command=su.IrMCSyncCommand,
    headset=su.Headset,
    audio_source=su.AudioSource,
    audio_sink=su.AudioSink,
    av_remove_control_target=su.AV_RemoteControlTarget,
    advanced_audio_distribution=su.AdvancedAudioDistribution,
    av_remote_control=su.AV_RemoteControl,
    av_remote_control_controller=su.AV_RemoteControlController,
    headset_ag=su.HeadsetAG,
    panu=su.PANU,
    nap=su.NAP,
    gn=su.GN,
    direct_printing=su.DirectPrinting,
    reference_printing=su.ReferencePrinting,
    basic_image=su.BasicImage,
    imaging_responder=su.ImagingResponder,
    imaging_automatic_archive=su.ImagingAutomaticArchive,
    imaging_reference_objects=su.ImagingReferenceObjects,
    hands_free=su.Handsfree,
    hands_free_audio_gateway=su.HandsfreeAudioGateway,
    direct_printing_reference_objects_service=su.DirectPrintingReferenceObjectsService,
    reflected_ui=su.ReflectedUI,
    basic_printing=su.BasicPrinting,
    printing_status=su.PrintingStatus,
    human_interface_device_service=su.HumanInterfaceDeviceService,
    hardcopy_cable_replacement=su.HardcopyCableReplacement,
    hcr_print=su.HCRPrint,
    hcr_scan=su.HCRScan,
    sim_access=su.SIMAccess,
    phonebook_access_pce=su.PhonebookAccessPCE,
    phonebook_access_pse=su.PhonebookAccessPSE,
    phonebook_access=su.PhonebookAccess,
    headset_hs=su.HeadsetHS,
    message_access_server=su.MessageAccessServer,
    message_notification_server=su.MessageNotificationServer,
    message_access_profile=su.MessageAccessProfile,
    gnss=su.GNSS,
    gnss_server=su.GNSSServer,
    display_3d=su.Display3D,
    glasses_3d=su.Glasses3D,
    synchronization_3d=su.Synchronization3D,
    mps_profile=su.MPSProfile,
    mps_service=su.MPSService,
    pnp_information=su.PnPInformation,
    generic_networking=su.GenericNetworking,
    generic_file_transfer=su.GenericFileTransfer,
    generic_audio=su.GenericAudio,
    generic_telephony=su.GenericTelephony,
    video_source=su.VideoSource,
    video_sink=su.VideoSink,
    video_distribution=su.VideoDistribution,
    hdp=su.HDP,
    hdp_source=su.HDPSource,
    hdp_sink=su.HDPSink,
    generic_access=su.GenericAccess,
    generic_attribute=su.GenericAttribute,
    immediate_alert=su.ImmediateAlert,
    link_loss=su.LinkLoss,
    tx_power=su.TxPower,
    current_time_service=su.CurrentTimeService,
    reference_time_update_service=su.ReferenceTimeUpdateService,
    next_dst_change_service=su.NextDSTChangeService,
    glucose=su.Glucose,
    health_thermometer=su.HealthThermometer,
    device_information=su.DeviceInformation,
    heart_rate=su.HeartRate,
    phone_alert_status_service=su.PhoneAlertStatusService,
    battery_service=su.BatteryService,
    blood_pressure=su.BloodPressure,
    alert_notification_service=su.AlertNotificationService,
    human_interface_device=su.HumanInterfaceDevice,
    scan_parameters=su.ScanParameters,
    running_speed_and_cadence=su.RunningSpeedAndCadence,
    cycling_speed_and_cadence=su.CyclingSpeedAndCadence,
    cycling_power=su.CyclingPower,
    location_and_navigation=su.LocationAndNavigation,
    environmental_sensing=su.EnvironmentalSensing,
    body_composition=su.BodyComposition,
    user_data=su.UserData,
    weight_scale=su.WeightScale,
    bond_management=su.BondManagement,
    continuous_glucose_monitoring=su.ContinuousGlucoseMonitoring,
)


class BluetoothUuid(core.UuidMixin, QtBluetooth.QBluetoothUuid):
    """Generates a UUID for each Bluetooth service."""

    @classmethod
    def characteristic_to_string(cls, characteristic: CharacteristicTypeStr | md) -> str:
        """Return string representation for given characteristic."""
        return cls.characteristicToString(
            CHARACTERISTIC_TYPES.get_enum_value(characteristic)
        )

    @classmethod
    def descriptor_to_string(cls, descriptor: DescriptorTypeStr | dt) -> str:
        """Return string representation for given descriptor."""
        return cls.descriptorToString(DESCRIPTOR_TYPE.get_enum_value(descriptor))

    @classmethod
    def protocol_to_string(cls, protocol: ProtocolUuidstr | pu) -> str:
        """Return string representation for given protocol."""
        return cls.protocolToString(PROTOCOL_UUID.get_enum_value(protocol))

    @classmethod
    def service_class_to_string(cls, service_class: DescriptorTypeStr | su) -> str:
        """Return string representation for given service class."""
        return cls.serviceClassToString(SERVICE_CLASS_UUID.get_enum_value(service_class))


if __name__ == "__main__":
    address = BluetoothUuid()
    print(BluetoothUuid.service_class_to_string("bond_management"))



"""Core non-graphical classes used by other modules."""

from __future__ import annotations

import contextlib
import sys
from prettyqt.qt import QtCore


from prettyqt.qt.QtCore import *  # noqa: F403

from prettyqt.qt.QtCore import (
    qInstallMessageHandler as install_message_handler,  # noqa: N813
    QtMsgType as MsgType,
    Signal,
    # Property,
    QEnum as Enum,
    QPoint as Point,
    QPointF as PointF,
    QModelIndex as ModelIndex,
    QRect as Rect,
    QRectF as RectF,
    # QtCriticalMsg as CriticalMsg,
    # QtDebugMsg as DebugMsg,
    # QtFatalMsg as FatalMsg,
    # QtInfoMsg as InfoMsg,
    # QtSystemMsg as SystemMsg,
    # QtWarningMsg as WarningMsg,
)

from .event import Event
from .slot import Slot

from .property import Property
from .loggingcategory import LoggingCategory
from .timezone import TimeZone
from .date import Date
from ._time import Time
from ._datetime import DateTime
from .semaphore import Semaphore
from .mutex import Mutex
from .metaenum import MetaEnum
from .metamethod import MetaMethod
from .metaproperty import MetaProperty
from .metatype import MetaType
from .metaobject import MetaObject
from .modelroledata import ModelRoleData
from .persistentmodelindex import PersistentModelIndex
from .cryptographichash import CryptographicHash
from ._uuid import Uuid, UuidMixin
from .signalblocker import SignalBlocker
from .abstractnativeeventfilter import AbstractNativeEventFilter

# from .debug import Debug
from .line import Line
from .linef import LineF
from .easingcurve import EasingCurve
from .processenvironment import ProcessEnvironment
from .randomgenerator import RandomGenerator

# from .randomgenerator64 import RandomGenerator64
from .deadlinetimer import DeadlineTimer
from .elapsedtimer import ElapsedTimer
from .basictimer import BasicTimer
from .object import Object, ObjectMixin
from .socketnotifier import SocketNotifier
from .sharedmemory import SharedMemory
from .signalmapper import SignalMapper
from .timeline import TimeLine
from .margins import Margins
from .marginsf import MarginsF
from ._locale import Locale
from .abstracteventdispatcher import AbstractEventDispatcher
from .textstream import TextStream
from .eventloop import EventLoop
from .abstractanimation import AbstractAnimation, AbstractAnimationMixin
from .variantanimation import VariantAnimation, VariantAnimationMixin
from .propertyanimation import PropertyAnimation
from .pauseanimation import PauseAnimation
from .animationgroup import AnimationGroup, AnimationGroupMixin
from .parallelanimationgroup import ParallelAnimationGroup
from .sequentialanimationgroup import SequentialAnimationGroup
from .versionnumber import VersionNumber
from .operatingsystemversion import OperatingSystemVersion
from .libraryinfo import LibraryInfo
from .datastream import DataStream
from .bytearray import ByteArray
from .bytearraymatcher import ByteArrayMatcher
from .url import Url
from .urlquery import UrlQuery
from .translator import Translator
from .coreapplication import CoreApplication, CoreApplicationMixin
from .filesystemwatcher import FileSystemWatcher
from .fileselector import FileSelector
from .iodevice import IODevice, IODeviceMixin
from .filedevice import FileDevice, FileDeviceMixin
from .file import File, FileMixin
from .savefile import SaveFile
from .lockfile import LockFile
from .buffer import Buffer
from .settings import Settings
from ._calendar import Calendar
from ._resource import Resource
from .size import Size
from .sizef import SizeF
from .fileinfo import FileInfo
from .storageinfo import StorageInfo

# from .point import Point
# from .pointf import PointF
from .textboundaryfinder import TextBoundaryFinder
from .timer import Timer
from .thread import Thread
from .process import Process

# from .rect import Rect
# from .rectf import RectF
from .regularexpressionmatch import RegularExpressionMatch
from .regularexpressionmatchiterator import RegularExpressionMatchIterator
from .regularexpression import RegularExpression
from .mimetype import MimeType
from .mimedatabase import MimeDatabase
from .mimedata import MimeData
from .runnable import Runnable

# from .modelindex import ModelIndex
from .threadpool import ThreadPool
from .dir import Dir
from .temporaryfile import TemporaryFile
from .temporarydir import TemporaryDir
from .keycombination import KeyCombination

# from .event import Event
from .itemselectionmodel import ItemSelectionModel
from .itemselection import ItemSelection
from .itemselectionrange import ItemSelectionRange
from .diriterator import DirIterator
from .abstractitemmodel import AbstractItemModel, AbstractItemModelMixin
from .abstractproxymodel import AbstractProxyModel, AbstractProxyModelMixin
from .identityproxymodel import IdentityProxyModel
from .abstractlistmodel import AbstractListModel, AbstractListModelMixin
from .stringlistmodel import StringListModel, StringListModelMixin
from .sortfilterproxymodel import SortFilterProxyModel
from .abstracttablemodel import AbstractTableModel, AbstractTableModelMixin
from .standardpaths import StandardPaths
from .xmlstreamreader import XmlStreamReader
from .commandlineoption import CommandLineOption
from .commandlineparser import CommandLineParser
from .collatorsortkey import CollatorSortKey
from .collator import Collator

from .jsonvalue import JsonValue
from .jsondocument import JsonDocument

from .library import Library  # type: ignore
from .pluginloader import PluginLoader  # type: ignore

from .concatenatetablesproxymodel import ConcatenateTablesProxyModel
from .transposeproxymodel import TransposeProxyModel

QT_MODULE = QtCore


@contextlib.contextmanager
def update_property_group():
    QtCore.Qt.beginPropertyUpdateGroup()
    yield None
    QtCore.Qt.endPropertyUpdateGroup()


def app(args: list[str] | None = None, **kwargs) -> CoreApplication:
    if (instance := CoreApplication.instance()) is not None:
        return instance
    return CoreApplication(sys.argv if args is None else args, **kwargs)


__all__ = [
    # "RandomGenerator64",
    "AbstractAnimation",
    "AbstractAnimationMixin",
    "AbstractEventDispatcher",
    # "SignalInstance",
    "AbstractItemModel",
    "AbstractItemModelMixin",
    "AbstractListModel",
    "AbstractListModelMixin",
    "AbstractNativeEventFilter",
    "AbstractProxyModel",
    "AbstractProxyModelMixin",
    "AbstractTableModel",
    "AbstractTableModelMixin",
    "AnimationGroup",
    "AnimationGroupMixin",
    "BasicTimer",
    "Buffer",
    "ByteArray",
    "ByteArrayMatcher",
    "Calendar",
    "Collator",
    "CollatorSortKey",
    # "AbstractState",
    # "FinalState",
    # "HistoryState",
    # "State",
    # "StateMachine",
    # "AbstractTransition",
    # "SignalTransition",
    # "EventTransition",
    "CommandLineOption",
    "CommandLineParser",
    "ConcatenateTablesProxyModel",
    "CoreApplication",
    "CoreApplicationMixin",
    "CryptographicHash",
    "DataStream",
    "Date",
    "DateTime",
    # "Debug",
    "DeadlineTimer",
    "Dir",
    "DirIterator",
    "EasingCurve",
    "ElapsedTimer",
    # "ChildEvent",
    # "TimerEvent",
    "Enum",
    "Event",
    "EventLoop",
    "File",
    "FileDevice",
    "FileDeviceMixin",
    "FileInfo",
    "FileMixin",
    "FileSelector",
    "FileSystemWatcher",
    "IODevice",
    "IODeviceMixin",
    "IdentityProxyModel",
    "ItemSelection",
    "ItemSelectionModel",
    "ItemSelectionRange",
    "JsonDocument",
    "JsonValue",
    "KeyCombination",
    "Library",
    "LibraryInfo",
    "Line",
    "LineF",
    "Locale",
    "LockFile",
    "LoggingCategory",
    "Margins",
    "MarginsF",
    # "ClassInfo",
    # "Q_ARG",
    # "Q_RETURN_ARG",
    # "DynamicPropertyChangeEvent",
    "MetaEnum",
    "MetaMethod",
    "MetaObject",
    "MetaProperty",
    "MetaType",
    "MimeData",
    "MimeDatabase",
    "MimeType",
    "ModelIndex",
    "ModelRoleData",
    "MsgType",
    "Mutex",
    "Object",
    "ObjectMixin",
    "OperatingSystemVersion",
    "ParallelAnimationGroup",
    "PauseAnimation",
    "PersistentModelIndex",
    "PersistentModelIndex",
    "PluginLoader",
    "Point",
    "PointF",
    "Process",
    "ProcessEnvironment",
    "Property",
    "PropertyAnimation",
    "RandomGenerator",
    "Rect",
    "RectF",
    "RegularExpression",
    "RegularExpressionMatch",
    "RegularExpressionMatchIterator",
    "Resource",
    "Runnable",
    "SaveFile",
    # "CriticalMsg",
    # "DebugMsg",
    # "FatalMsg",
    # "InfoMsg",
    # "SystemMsg",
    # "WarningMsg",
    "Semaphore",
    "SequentialAnimationGroup",
    "Settings",
    "SharedMemory",
    "Signal",
    "SignalBlocker",
    "SignalMapper",
    "Size",
    "SizeF",
    "Slot",
    "SocketNotifier",
    "SortFilterProxyModel",
    "StandardPaths",
    "StorageInfo",
    "StringListModel",
    "StringListModelMixin",
    "TemporaryDir",
    "TemporaryFile",
    "TextBoundaryFinder",
    "TextStream",
    "Thread",
    "ThreadPool",
    "Time",
    "TimeLine",
    "TimeZone",
    "Timer",
    "Translator",
    "TransposeProxyModel",
    "Url",
    "UrlQuery",
    "Uuid",
    "UuidMixin",
    "VariantAnimation",
    "VariantAnimationMixin",
    "VersionNumber",
    "XmlStreamReader",
    "app",
    "install_message_handler",
]



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.qt import QtGui


class StyleHintsMixin(core.ObjectMixin):
    def get_color_scheme(self) -> constants.ColorSchemeStr:
        """Return current color scheme of the system.

        Returns:
            color scheme
        """
        return constants.COLOR_SCHEME.inverse[self.colorScheme()]

    def get_tab_focus_behavior(self) -> constants.TabFocusBehaviorStr:
        """Return current focus behavior on press of the tab key.

        Returns:
            focus behavior
        """
        return constants.TAB_FOCUS_BEHAVIOR.inverse[self.tabFocusBehavior()]


class StyleHintsType(type):
    def __getattr__(cls, key):
        return getattr(QtGui.QStyleHints, key)


class StyleHints(StyleHintsMixin, metaclass=StyleHintsType):
    def __init__(self, item: QtGui.QStyleHints):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    hints = StyleHints(app.styleHints())
    scheme = hints.get_color_scheme()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui, widgets


if TYPE_CHECKING:
    from prettyqt import core


class ClickableLabel(widgets.Label):
    """A label widget that behaves like a button."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, alignment="left")
        self.setFont(gui.QFont("Arial"))
        self.setFixedHeight(24)
        self.set_size_policy("minimum", "expanding")

    def setText(self, text: str):
        fm = gui.FontMetrics(self.font())
        width = fm.horizontalAdvance(text)
        self.setFixedWidth(width + 18)
        super().setText(text)

    def enterEvent(self, event: core.QEvent):
        with self.edit_font() as font:
            font.setUnderline(True)
        self.set_cursor("pointing_hand")
        self.update()
        return super().enterEvent(event)

    def leaveEvent(self, event: core.QEvent):
        with self.edit_font() as font:
            font.setUnderline(False)
        self.unsetCursor()
        return super().leaveEvent(event)


if __name__ == "__main__":
    app = widgets.app()
    widget = ClickableLabel("test", tool_tip="testus")
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr


mod = QtCore.QEvent

TypeStr = Literal[
    "none",
    "action_added",
    "action_changed",
    "action_removed",
    "activation_change",
    "application_activate",
    "application_deactivate",
    "application_font_change",
    "application_layout_direction_change",
    "application_palette_change",
    "application_state_change",
    "application_window_icon_change",
    "child_added",
    "child_polished",
    "child_removed",
    "clipboard",
    "close",
    "close_software_input_panel",
    "contents_rect_change",
    "context_menu",
    "cursor_change",
    "deferred_delete",
    "drag_enter",
    "drag_leave",
    "drag_move",
    "drop",
    "dynamic_property_change",
    "enabled_change",
    "enter",
    "enter_whats_this_mode",
    "expose",
    "file_open",
    "focus_in",
    "focus_out",
    "focus_about_to_change",
    "font_change",
    "gesture",
    "gesture_override",
    "grab_keyboard",
    "grab_mouse",
    "graphics_scene_context_menu",
    "graphics_scene_drag_enter",
    "graphics_scene_drag_leave",
    "graphics_scene_drag_move",
    "graphics_scene_drop",
    "graphics_scene_help",
    "graphics_scene_hover_enter",
    "graphics_scene_hover_leave",
    "graphics_scene_hover_move",
    "graphics_scene_mouse_double_click",
    "graphics_scene_mouse_move",
    "graphics_scene_mouse_press",
    "graphics_scene_mouse_release",
    "graphics_scene_move",
    "graphics_scene_resize",
    "graphics_scene_wheel",
    "hide",
    "hide_to_parent",
    "hover_enter",
    "hover_leave",
    "hover_move",
    "icon_drag",
    "icon_text_change",
    "input_method",
    "input_method_query",
    "keyboard_layout_change",
    "key_press",
    "key_release",
    "language_change",
    "layout_direction_change",
    "layout_request",
    "leave",
    "leave_whats_this_mode",
    "locale_change",
    "non_client_area_mouse_button_dbl_click",
    "non_client_area_mouse_button_press",
    "non_client_area_mouse_button_release",
    "non_client_area_mouse_move",
    "mac_size_change",
    "meta_call",
    "modified_change",
    "mouse_button_dbl_click",
    "mouse_button_press",
    "mouse_button_release",
    "mouse_move",
    "mouse_tracking_change",
    "move",
    "native_gesture",
    "orientation_change",
    "paint",
    "palette_change",
    "parent_about_to_change",
    "parent_change",
    "platform_panel",
    "platform_surface",
    "polish",
    "polish_request",
    "query_whats_this",
    "quit",
    "read_only_change",
    "request_software_input_panel",
    "resize",
    "scroll_prepare",
    "scroll",
    "shortcut",
    "shortcut_override",
    "show",
    "show_to_parent",
    "sock_act",
    "state_machine_signal",
    "state_machine_wrapped",
    "status_tip",
    "style_change",
    "tablet_move",
    "tablet_press",
    "tablet_release",
    "tablet_enter_proximity",
    "tablet_leave_proximity",
    "tablet_tracking_change",
    "thread_change",
    "timer",
    "tool_bar_change",
    "tool_tip",
    "tool_tip_change",
    "touch_begin",
    "touch_cancel",
    "touch_end",
    "touch_update",
    "ungrab_keyboard",
    "ungrab_mouse",
    "update_later",
    "update_request",
    "whats_this",
    "whats_this_clicked",
    "wheel",
    "win_event_act",
    "window_activate",
    "window_blocked",
    "window_deactivate",
    "window_icon_change",
    "window_state_change",
    "window_title_change",
    "window_unblocked",
    "win_id_change",
    "z_order_change",
]


TYPE: bidict[TypeStr, mod.Type] = bidict(
    none=mod.Type.None_,
    action_added=mod.Type.ActionAdded,
    action_changed=mod.Type.ActionChanged,
    action_removed=mod.Type.ActionRemoved,
    activation_change=mod.Type.ActivationChange,
    application_activate=mod.Type.ApplicationActivate,
    application_deactivate=mod.Type.ApplicationDeactivate,
    application_font_change=mod.Type.ApplicationFontChange,
    application_layout_direction_change=mod.Type.ApplicationLayoutDirectionChange,
    application_palette_change=mod.Type.ApplicationPaletteChange,
    application_state_change=mod.Type.ApplicationStateChange,
    application_window_icon_change=mod.Type.ApplicationWindowIconChange,
    child_added=mod.Type.ChildAdded,
    child_polished=mod.Type.ChildPolished,
    child_removed=mod.Type.ChildRemoved,
    clipboard=mod.Type.Clipboard,
    close=mod.Type.Close,
    close_software_input_panel=mod.Type.CloseSoftwareInputPanel,
    contents_rect_change=mod.Type.ContentsRectChange,
    context_menu=mod.Type.ContextMenu,
    cursor_change=mod.Type.CursorChange,
    deferred_delete=mod.Type.DeferredDelete,
    drag_enter=mod.Type.DragEnter,
    drag_leave=mod.Type.DragLeave,
    drag_move=mod.Type.DragMove,
    drop=mod.Type.Drop,
    dynamic_property_change=mod.Type.DynamicPropertyChange,
    enabled_change=mod.Type.EnabledChange,
    enter=mod.Type.Enter,
    # enter_edit_focus=mod.Type.EnterEditFocus,
    # leave_edit_focus=mod.Type.LeaveEditFocus ,
    enter_whats_this_mode=mod.Type.EnterWhatsThisMode,
    expose=mod.Type.Expose,
    file_open=mod.Type.FileOpen,
    focus_in=mod.Type.FocusIn,
    focus_out=mod.Type.FocusOut,
    focus_about_to_change=mod.Type.FocusAboutToChange,
    font_change=mod.Type.FontChange,
    gesture=mod.Type.Gesture,
    gesture_override=mod.Type.GestureOverride,
    grab_keyboard=mod.Type.GrabKeyboard,
    grab_mouse=mod.Type.GrabMouse,
    graphics_scene_context_menu=mod.Type.GraphicsSceneContextMenu,
    graphics_scene_drag_enter=mod.Type.GraphicsSceneDragEnter,
    graphics_scene_drag_leave=mod.Type.GraphicsSceneDragLeave,
    graphics_scene_drag_move=mod.Type.GraphicsSceneDragMove,
    graphics_scene_drop=mod.Type.GraphicsSceneDrop,
    graphics_scene_help=mod.Type.GraphicsSceneHelp,
    graphics_scene_hover_enter=mod.Type.GraphicsSceneHoverEnter,
    graphics_scene_hover_leave=mod.Type.GraphicsSceneHoverLeave,
    graphics_scene_hover_move=mod.Type.GraphicsSceneHoverMove,
    graphics_scene_mouse_double_click=mod.Type.GraphicsSceneMouseDoubleClick,
    graphics_scene_mouse_move=mod.Type.GraphicsSceneMouseMove,
    graphics_scene_mouse_press=mod.Type.GraphicsSceneMousePress,
    graphics_scene_mouse_release=mod.Type.GraphicsSceneMouseRelease,
    graphics_scene_move=mod.Type.GraphicsSceneMove,
    graphics_scene_resize=mod.Type.GraphicsSceneResize,
    graphics_scene_wheel=mod.Type.GraphicsSceneWheel,
    hide=mod.Type.Hide,
    hide_to_parent=mod.Type.HideToParent,
    hover_enter=mod.Type.HoverEnter,
    hover_leave=mod.Type.HoverLeave,
    hover_move=mod.Type.HoverMove,
    icon_drag=mod.Type.IconDrag,
    icon_text_change=mod.Type.IconTextChange,
    input_method=mod.Type.InputMethod,
    input_method_query=mod.Type.InputMethodQuery,
    keyboard_layout_change=mod.Type.KeyboardLayoutChange,
    key_press=mod.Type.KeyPress,
    key_release=mod.Type.KeyRelease,
    language_change=mod.Type.LanguageChange,
    layout_direction_change=mod.Type.LayoutDirectionChange,
    layout_request=mod.Type.LayoutRequest,
    leave=mod.Type.Leave,
    leave_whats_this_mode=mod.Type.LeaveWhatsThisMode,
    locale_change=mod.Type.LocaleChange,
    non_client_area_mouse_button_dbl_click=mod.Type.NonClientAreaMouseButtonDblClick,
    non_client_area_mouse_button_press=mod.Type.NonClientAreaMouseButtonPress,
    non_client_area_mouse_button_release=mod.Type.NonClientAreaMouseButtonRelease,
    non_client_area_mouse_move=mod.Type.NonClientAreaMouseMove,
    mac_size_change=mod.Type.MacSizeChange,
    meta_call=mod.Type.MetaCall,
    modified_change=mod.Type.ModifiedChange,
    mouse_button_dbl_click=mod.Type.MouseButtonDblClick,
    mouse_button_press=mod.Type.MouseButtonPress,
    mouse_button_release=mod.Type.MouseButtonRelease,
    mouse_move=mod.Type.MouseMove,
    mouse_tracking_change=mod.Type.MouseTrackingChange,
    move=mod.Type.Move,
    native_gesture=mod.Type.NativeGesture,
    orientation_change=mod.Type.OrientationChange,
    paint=mod.Type.Paint,
    palette_change=mod.Type.PaletteChange,
    parent_about_to_change=mod.Type.ParentAboutToChange,
    parent_change=mod.Type.ParentChange,
    platform_panel=mod.Type.PlatformPanel,
    platform_surface=mod.Type.PlatformSurface,
    polish=mod.Type.Polish,
    polish_request=mod.Type.PolishRequest,
    query_whats_this=mod.Type.QueryWhatsThis,
    quit=mod.Type.Quit,
    read_only_change=mod.Type.ReadOnlyChange,
    request_software_input_panel=mod.Type.RequestSoftwareInputPanel,
    resize=mod.Type.Resize,
    scroll_prepare=mod.Type.ScrollPrepare,
    scroll=mod.Type.Scroll,
    shortcut=mod.Type.Shortcut,
    shortcut_override=mod.Type.ShortcutOverride,
    show=mod.Type.Show,
    show_to_parent=mod.Type.ShowToParent,
    sock_act=mod.Type.SockAct,
    state_machine_signal=mod.Type.StateMachineSignal,
    state_machine_wrapped=mod.Type.StateMachineWrapped,
    status_tip=mod.Type.StatusTip,
    style_change=mod.Type.StyleChange,
    tablet_move=mod.Type.TabletMove,
    tablet_press=mod.Type.TabletPress,
    tablet_release=mod.Type.TabletRelease,
    tablet_enter_proximity=mod.Type.TabletEnterProximity,
    tablet_leave_proximity=mod.Type.TabletLeaveProximity,
    tablet_tracking_change=mod.Type.TabletTrackingChange,
    thread_change=mod.Type.ThreadChange,
    timer=mod.Type.Timer,
    tool_bar_change=mod.Type.ToolBarChange,
    tool_tip=mod.Type.ToolTip,
    tool_tip_change=mod.Type.ToolTipChange,
    touch_begin=mod.Type.TouchBegin,
    touch_cancel=mod.Type.TouchCancel,
    touch_end=mod.Type.TouchEnd,
    touch_update=mod.Type.TouchUpdate,
    ungrab_keyboard=mod.Type.UngrabKeyboard,
    ungrab_mouse=mod.Type.UngrabMouse,
    update_later=mod.Type.UpdateLater,
    update_request=mod.Type.UpdateRequest,
    whats_this=mod.Type.WhatsThis,
    whats_this_clicked=mod.Type.WhatsThisClicked,
    wheel=mod.Type.Wheel,
    win_event_act=mod.Type.WinEventAct,
    window_activate=mod.Type.WindowActivate,
    window_blocked=mod.Type.WindowBlocked,
    window_deactivate=mod.Type.WindowDeactivate,
    window_icon_change=mod.Type.WindowIconChange,
    window_state_change=mod.Type.WindowStateChange,
    window_title_change=mod.Type.WindowTitleChange,
    window_unblocked=mod.Type.WindowUnblocked,
    win_id_change=mod.Type.WinIdChange,
    z_order_change=mod.Type.ZOrderChange,
)

MouseButtonTypeStr = Literal["press", "release", "double_click", "move"]

MouseButtonType: bidict[MouseButtonTypeStr, QtCore.QEvent.Type] = bidict(
    press=QtCore.QEvent.Type.MouseButtonPress,
    release=QtCore.QEvent.Type.MouseButtonRelease,
    double_click=QtCore.QEvent.Type.MouseButtonDblClick,
    move=QtCore.QEvent.Type.MouseMove,
)


class Event(QtCore.QEvent):
    """The base class of all event classes. Event objects contain event parameters."""

    def __repr__(self):
        return get_repr(self, self.type())



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core


if TYPE_CHECKING:
    from collections.abc import Iterator


class ItemSelectionRange(core.QItemSelectionRange):
    """Manages information about a range of selected items in a model."""

    def __contains__(self, other: core.ModelIndex):
        return self.contains(other)

    def __bool__(self):
        return not self.isEmpty()

    def __iter__(self) -> Iterator[core.ModelIndex]:
        return iter(self.indexes())  # type: ignore

    def __len__(self):
        return len(self.indexes())

    def __and__(self, other: core.QItemSelectionRange):
        return self.intersected(other)


if __name__ == "__main__":
    temp = ItemSelectionRange()
    temp2 = ItemSelectionRange()
    temp3 = temp & temp2



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import gui
from prettyqt.utils import bidict, colors


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


mod = gui.QTextCharFormat

FontPropertiesInheritanceBehaviorStr = Literal["none", "single"]

FONT_PROPERTY_INHERITANCE_BEHAVIOUR: bidict[
    FontPropertiesInheritanceBehaviorStr, mod.FontPropertiesInheritanceBehavior
] = bidict(
    none=mod.FontPropertiesInheritanceBehavior.FontPropertiesSpecifiedOnly,
    single=mod.FontPropertiesInheritanceBehavior.FontPropertiesAll,
)

UnderlineStyleStr = Literal[
    "none",
    "single",
    "dash",
    "dot",
    "dashdot",
    "dashdotline",
    "wave",
    "spellcheck",
]

UNDERLINE_STYLE: bidict[UnderlineStyleStr, mod.UnderlineStyle] = bidict(
    none=mod.UnderlineStyle.NoUnderline,
    single=mod.UnderlineStyle.SingleUnderline,
    dash=mod.UnderlineStyle.DashUnderline,
    dot=mod.UnderlineStyle.DotLine,
    dashdot=mod.UnderlineStyle.DashDotLine,
    dashdotline=mod.UnderlineStyle.DashDotDotLine,
    wave=mod.UnderlineStyle.WaveUnderline,
    spellcheck=mod.UnderlineStyle.SpellCheckUnderline,
)

VerticalAlignmentStr = Literal[
    "normal",
    "super_script",
    "sub_script",
    "middle",
    "bottom",
    "top",
    "baseline",
]

VERTICAL_ALIGNMENT: bidict[
    VerticalAlignmentStr, gui.QTextCharFormat.VerticalAlignment
] = bidict(
    normal=gui.QTextCharFormat.VerticalAlignment.AlignNormal,
    super_script=gui.QTextCharFormat.VerticalAlignment.AlignSuperScript,
    sub_script=gui.QTextCharFormat.VerticalAlignment.AlignSubScript,
    middle=gui.QTextCharFormat.VerticalAlignment.AlignMiddle,
    bottom=gui.QTextCharFormat.VerticalAlignment.AlignBottom,
    top=gui.QTextCharFormat.VerticalAlignment.AlignTop,
    baseline=gui.QTextCharFormat.VerticalAlignment.AlignBaseline,
)


class TextCharFormatMixin(gui.TextFormatMixin):
    def __init__(
        self,
        text_color: datatypes.ColorType | gui.QBrush = None,
        bold: bool = False,
        italic: bool = False,
    ):
        super().__init__()
        if text_color is not None:
            self.set_foreground_color(text_color)
        if bold:
            self.set_font_weight("bold")
        self.setFontItalic(italic)

    def set_foreground_color(self, color: datatypes.ColorType | gui.QBrush):
        if not isinstance(color, gui.QBrush):
            color = colors.get_color(color)
        self.setForeground(color)

    def set_background_color(self, color: datatypes.ColorType | gui.QBrush):
        if not isinstance(color, gui.QBrush):
            color = colors.get_color(color)
        self.setBackground(color)

    def set_font_weight(self, weight: gui.font.WeightStr | gui.Font.Weight):
        """Set the font weight.

        Args:
            weight: font weight
        """
        self.setFontWeight(gui.font.WEIGHT.get_enum_value(weight))

    def get_font_weight(self) -> gui.font.WeightStr:
        """Get current font weight.

        Returns:
            current font weight
        """
        return gui.font.WEIGHT.inverse[self.fontWeight()]

    def set_underline_style(self, style: UnderlineStyleStr | mod.UnderlineStyle):
        """Set the underline style.

        Args:
            style: underline style
        """
        self.setUnderlineStyle(UNDERLINE_STYLE.get_enum_value(style))

    def get_underline_style(self) -> UnderlineStyleStr:
        """Get current underline style.

        Returns:
            current underline style
        """
        return UNDERLINE_STYLE.inverse[self.underlineStyle()]

    def set_vertical_alignment(
        self, alignment: VerticalAlignmentStr | gui.QTextCharFormat.VerticalAlignment
    ):
        """Set the vertical alignment.

        Args:
            alignment: vertical alignment
        """
        self.setVerticalAlignment(VERTICAL_ALIGNMENT.get_enum_value(alignment))

    def get_vertical_alignment(self) -> VerticalAlignmentStr:
        """Get current vertical alignment.

        Returns:
            current vertical alignment
        """
        return VERTICAL_ALIGNMENT.inverse[self.verticalAlignment()]

    def set_font_style_hint(self, hint: gui.font.StyleHintStr | gui.Font.StyleHint):
        """Set the font style hint.

        Args:
            hint: font style hint
        """
        self.setFontStyleHint(gui.font.STYLE_HINTS.get_enum_value(hint))

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())


class TextCharFormat(TextCharFormatMixin, gui.QTextCharFormat):
    pass


if __name__ == "__main__":
    w = TextCharFormat()



from __future__ import annotations

from typing import Literal

from prettyqt import core, widgets
from prettyqt.utils import bidict


ModeStr = Literal["hex", "decimal", "octal", "binary"]

MODE: bidict[ModeStr, widgets.QLCDNumber.Mode] = bidict(
    hex=widgets.QLCDNumber.Mode.Hex,
    decimal=widgets.QLCDNumber.Mode.Dec,
    octal=widgets.QLCDNumber.Mode.Oct,
    binary=widgets.QLCDNumber.Mode.Bin,
)

SegmentStyleStr = Literal["outline", "filled", "flat"]

SEGMENT_STYLE: bidict[SegmentStyleStr, widgets.QLCDNumber.SegmentStyle] = bidict(
    outline=widgets.QLCDNumber.SegmentStyle.Outline,
    filled=widgets.QLCDNumber.SegmentStyle.Filled,
    flat=widgets.QLCDNumber.SegmentStyle.Flat,
)


class LCDNumber(widgets.FrameMixin, widgets.QLCDNumber):
    """Displays a number with LCD-like digits."""

    value_changed = core.Signal(float)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"segmentStyle": SEGMENT_STYLE, "mode": MODE}
        return maps

    def set_mode(self, mode: ModeStr | widgets.QLCDNumber.Mode):
        """Set the lcd mode.

        Args:
            mode: lcd mode to use
        """
        self.setMode(MODE.get_enum_value(mode))

    def get_mode(self) -> ModeStr:
        """Return current lcd mode.

        Returns:
            lcd mode
        """
        return MODE.inverse[self.mode()]

    def set_segment_style(self, style: SegmentStyleStr | widgets.QLCDNumber.SegmentStyle):
        """Set the segment style.

        Args:
            style: segment style to use
        """
        self.setSegmentStyle(SEGMENT_STYLE.get_enum_value(style))

    def get_segment_style(self) -> SegmentStyleStr:
        """Return current segment style.

        Returns:
            segment style
        """
        return SEGMENT_STYLE.inverse[self.segmentStyle()]

    def set_value(self, value: float | str):
        if value != self.value():
            self.value_changed.emit(float(value))
        self.display(value)

    def get_value(self) -> float:
        return self.value()


if __name__ == "__main__":
    app = widgets.app()
    widget = LCDNumber()
    widget.set_value(5555)
    widget.show()
    app.exec()



from __future__ import annotations

import logging

import mknodes

from prettyqt import custom_widgets, prettyqtmarkdown, widgets
from prettyqt.utils import helpers


logger = logging.getLogger(__name__)


class MarkdownWidget(widgets.Widget):
    """A widget used for displaying a preview of a generated mknodes tree."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = widgets.TreeView()
        self.textbox = custom_widgets.CodeEditor()
        self.textbox.set_syntaxhighlighter("markdown")
        layout = self.set_layout("grid")
        layout[0, 0] = self.tree
        layout[0, 1] = self.textbox

    def set_markdown(self, item: mknodes.MkNode):
        model = prettyqtmarkdown.MkNodesModel(item)
        self.tree.set_model(model)
        self.tree.selectionModel().currentRowChanged.connect(self._on_current_change)

    def _on_current_change(self, new, old):
        text = new.data(new.model().Roles.MarkdownRole)
        self.textbox.set_text(text)


if __name__ == "__main__":
    app = widgets.app()
    page = mknodes.MkPage()
    page += mknodes.MkAdmonition("test")
    page += mknodes.MkTable(data=dict(a=[1, 2], b=["c", "D"]), header="From mapping")
    page += mknodes.MkDocStrings(helpers, header="DocStrings")
    widget = MarkdownWidget()
    widget.set_markdown(page)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import qml
from prettyqt.qt import QtQuick


class QuickImageProviderMixin(qml.QmlImageProviderBaseMixin):
    pass


class QuickImageProvider(QuickImageProviderMixin, QtQuick.QQuickImageProvider):
    """Interface for supporting pixmaps and threaded image requests in QML."""



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui, qml
from prettyqt.qt import QtQuick
from prettyqt.utils import bidict, datatypes


FlagStr = Literal[
    "clips_children_to_shape",
    "accepts_input_method",
    "is_focus_scope",
    "has_contents",
    "accepts_drops",
]

FLAGS: bidict[FlagStr, QtQuick.QQuickItem.Flag] = bidict(
    clips_children_to_shape=QtQuick.QQuickItem.Flag.ItemClipsChildrenToShape,
    accepts_input_method=QtQuick.QQuickItem.Flag.ItemAcceptsInputMethod,
    is_focus_scope=QtQuick.QQuickItem.Flag.ItemIsFocusScope,
    has_contents=QtQuick.QQuickItem.Flag.ItemHasContents,
    accepts_drops=QtQuick.QQuickItem.Flag.ItemAcceptsDrops,
)

ItemChangeStr = Literal[
    "child_added_change",
    "child_removed_change",
    "item_scene_change",
    "visible_has_changed",
    "parent_has_changed",
    "opacity_has_changed",
    "active_focus_has_changed",
    "rotation_has_changed",
    "pixel_ratio_has_changed",
    "anti_aliasing_has_changed",
    "enabled_has_changed",
]

ITEM_CHANGE: bidict[ItemChangeStr, QtQuick.QQuickItem.ItemChange] = bidict(
    child_added_change=QtQuick.QQuickItem.ItemChange.ItemChildAddedChange,
    child_removed_change=QtQuick.QQuickItem.ItemChange.ItemChildRemovedChange,
    item_scene_change=QtQuick.QQuickItem.ItemChange.ItemSceneChange,
    visible_has_changed=QtQuick.QQuickItem.ItemChange.ItemVisibleHasChanged,
    parent_has_changed=QtQuick.QQuickItem.ItemChange.ItemParentHasChanged,
    opacity_has_changed=QtQuick.QQuickItem.ItemChange.ItemOpacityHasChanged,
    active_focus_has_changed=QtQuick.QQuickItem.ItemChange.ItemActiveFocusHasChanged,
    rotation_has_changed=QtQuick.QQuickItem.ItemChange.ItemRotationHasChanged,
    pixel_ratio_has_changed=QtQuick.QQuickItem.ItemChange.ItemDevicePixelRatioHasChanged,
    anti_aliasing_has_changed=QtQuick.QQuickItem.ItemChange.ItemAntialiasingHasChanged,
    enabled_has_changed=QtQuick.QQuickItem.ItemChange.ItemEnabledHasChanged,
)

TransformOriginStr = Literal[
    "top_left",
    "top",
    "top_right",
    "left",
    "center",
    "right",
    "bottom_left",
    "bottom",
    "bottom_right",
]

TRANSFORM_ORIGIN: bidict[TransformOriginStr, QtQuick.QQuickItem.TransformOrigin] = bidict(
    top_left=QtQuick.QQuickItem.TransformOrigin.TopLeft,
    top=QtQuick.QQuickItem.TransformOrigin.Top,
    top_right=QtQuick.QQuickItem.TransformOrigin.TopRight,
    left=QtQuick.QQuickItem.TransformOrigin.Left,
    center=QtQuick.QQuickItem.TransformOrigin.Center,
    right=QtQuick.QQuickItem.TransformOrigin.Right,
    bottom_left=QtQuick.QQuickItem.TransformOrigin.BottomLeft,
    bottom=QtQuick.QQuickItem.TransformOrigin.Bottom,
    bottom_right=QtQuick.QQuickItem.TransformOrigin.BottomRight,
)


class QuickItemMixin(core.ObjectMixin, qml.QmlParserStatusMixin):
    def __getitem__(self, index: tuple[int, int]) -> QtQuick.QQuickItem:
        return self.childAt(*index)

    def __contains__(self, item: core.QPointF):
        return self.contains(item)

    def get_children_rect(self) -> core.RectF:
        return core.RectF(self.childrenRect())

    def get_cursor(self) -> gui.Cursor:
        return gui.Cursor(self.cursor())

    def get_flags(self):
        pass

    def set_transform_origin(
        self, origin: TransformOriginStr | QtQuick.QQuickItem.TransformOrigin
    ):
        """Set the origin point around which scale and rotation transform.

        The default is "center".

        Args:
            origin: transform origin to use
        """
        self.setTransformOrigin(TRANSFORM_ORIGIN.get_enum_value(origin))

    def get_transform_origin(self) -> TransformOriginStr:
        """Return the render type of text-like elements in Qt Quick.

        Returns:
            transform origin
        """
        return TRANSFORM_ORIGIN.inverse[self.transformOrigin()]

    def set_size(self, size: core.QSize | datatypes.SizeFType):
        self.setSize(datatypes.to_sizef(size))


class QuickItem(QuickItemMixin, QtQuick.QQuickItem):
    """The most basic of all visual items in Qt Quick."""


if __name__ == "__main__":
    item = QuickItem()



from __future__ import annotations

import datetime
import logging

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class SliceDisplayTextProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to convert and format non-str values for the DisplayRole.

    Usually, formatting of numbers etc is done by the ItemDelegate.
    By moving the formatting into a proxy layer instead, we can keep the ItemDelegate
    spot free for other stuff.

    Information about string formatting:

    https://docs.python.org/3/library/string.html#format-specification-mini-language

    ### Example

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table[:, :3].proxify.format_text(int_format="{:0>2d}")
    table.show()
    # or
    indexer = (slice(None), slice(None, 3))
    proxy = SliceDisplayTextProxyModel(indexer=indexer, float_format="{:.4f}")
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    def __init__(self, *args, **kwargs):
        self._int_format = "{:.4f}"
        self._float_format = "{:.4f}"
        self._datetime_format = "%m/%d/%Y, %H:%M:%S"
        self._date_format = "%m/%d/%Y"
        self._time_format = "%H:%M:%S"
        # self._force_override = False
        super().__init__(*args, **kwargs)

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not self.indexer_contains(index) or role != constants.DISPLAY_ROLE:
            return super().data(index, role)
        data = index.data()
        match data:
            case int():
                return self._int_format.format(data)
            case float():
                return self._float_format.format(data)
            case core.QDateTime():
                return data.toString(self._datetime_format)
            case datetime.datetime():
                return data.strftime(self._datetime_format)
            case core.QTime():
                return data.toString(self._time_format)
            case datetime.time():
                return data.strftime(self._time_format)
            case core.QDate():
                return data.toString(self._date_format)
            case datetime.date():
                return data.strftime(self._date_format)
            case _:
                return super().data(index, role)

    def set_int_format(self, fmt: str):
        self._int_format = fmt

    def get_int_format(self) -> str:
        return self._int_format

    def set_float_format(self, fmt: str):
        self._float_format = fmt

    def get_float_format(self) -> str:
        return self._float_format

    def set_datetime_format(self, fmt: str):
        self._datetime_format = fmt

    def get_datetime_format(self, fmt) -> str:
        return self._datetime_format

    def set_date_format(self, fmt: str):
        self._date_format = fmt

    def get_date_format(self, fmt) -> str:
        return self._date_format

    def set_time_format(self, fmt: str):
        self._time_format = fmt

    def get_time_format(self, fmt) -> str:
        return self._time_format

    int_format = core.Property(
        str,
        get_int_format,
        set_int_format,
        doc="String format for integers",
    )
    float_format = core.Property(
        str,
        get_float_format,
        set_float_format,
        doc="String format for floats",
    )
    datetime_format = core.Property(
        str,
        get_datetime_format,
        set_datetime_format,
        doc="String format for datetime objects",
    )
    date_format = core.Property(
        str,
        get_date_format,
        set_date_format,
        doc="String format for date objects",
    )
    time_format = core.Property(
        str,
        get_time_format,
        set_time_format,
        doc="String format for time objects",
    )


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()

    table = debugging.example_table()
    table.proxifier[1:4, :].set_format()
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict


Item = QtWebEngineCore.QWebEngineDownloadRequest

DownloadInterruptReasonStr = Literal[
    "none",
    "file_failed",
    "file_access_denied",
    "file_no_space",
    "filename_too_long",
    "file_too_large",
    "file_virus_infected",
    "file_transient_error",
    "file_blocked",
    "file_security_check_failed",
    "file_too_short",
    "file_hash_mismatch",
    "network_fialed",
    "network_timeout",
    "network_disconnected",
    "network_server_down",
    "network_invalid_request",
    "server_failed",
    "server_bad_content",
    "server_unauthorized",
    "server_cert_problem",
    "server_forbidden",
    "server_unreachable",
    "user_canceled",
]

DOWNLOAD_INTERRUPT_REASONS: bidict[
    DownloadInterruptReasonStr, Item.DownloadInterruptReason
] = bidict(
    none=Item.DownloadInterruptReason.NoReason,
    file_failed=Item.DownloadInterruptReason.FileFailed,
    file_access_denied=Item.DownloadInterruptReason.FileAccessDenied,
    file_no_space=Item.DownloadInterruptReason.FileNoSpace,
    filename_too_long=Item.DownloadInterruptReason.FileNameTooLong,
    file_too_large=Item.DownloadInterruptReason.FileTooLarge,
    file_virus_infected=Item.DownloadInterruptReason.FileVirusInfected,
    file_transient_error=Item.DownloadInterruptReason.FileTransientError,
    file_blocked=Item.DownloadInterruptReason.FileBlocked,
    file_security_check_failed=Item.DownloadInterruptReason.FileSecurityCheckFailed,
    file_too_short=Item.DownloadInterruptReason.FileTooShort,
    file_hash_mismatch=Item.DownloadInterruptReason.FileHashMismatch,
    network_fialed=Item.DownloadInterruptReason.NetworkFailed,
    network_timeout=Item.DownloadInterruptReason.NetworkTimeout,
    network_disconnected=Item.DownloadInterruptReason.NetworkDisconnected,
    network_server_down=Item.DownloadInterruptReason.NetworkServerDown,
    network_invalid_request=Item.DownloadInterruptReason.NetworkInvalidRequest,
    server_failed=Item.DownloadInterruptReason.ServerFailed,
    server_bad_content=Item.DownloadInterruptReason.ServerBadContent,
    server_unauthorized=Item.DownloadInterruptReason.ServerUnauthorized,
    server_cert_problem=Item.DownloadInterruptReason.ServerCertProblem,
    server_forbidden=Item.DownloadInterruptReason.ServerForbidden,
    server_unreachable=Item.DownloadInterruptReason.ServerUnreachable,
    user_canceled=Item.DownloadInterruptReason.UserCanceled,
)

DownloadStateStr = Literal[
    "requested", "in_progress", "completed", "cancelled", "interrupted"
]

DOWNLOAD_STATE: bidict[DownloadStateStr, Item.DownloadState] = bidict(
    requested=Item.DownloadState.DownloadRequested,
    in_progress=Item.DownloadState.DownloadInProgress,
    completed=Item.DownloadState.DownloadCompleted,
    cancelled=Item.DownloadState.DownloadCancelled,
    interrupted=Item.DownloadState.DownloadInterrupted,
)

SavePageFormatStr = Literal["unknown", "single_html", "complete_html", "mime_html"]

SAVE_PAGE_FORMAT: bidict[SavePageFormatStr, Item.SavePageFormat] = bidict(
    unknown=Item.SavePageFormat.UnknownSaveFormat,
    single_html=Item.SavePageFormat.SingleHtmlSaveFormat,
    complete_html=Item.SavePageFormat.CompleteHtmlSaveFormat,
    mime_html=Item.SavePageFormat.MimeHtmlSaveFormat,
)

# Item.__bases__ = (core.Object,)


class WebEngineDownloadRequest:
    def __init__(self, item: QtWebEngineCore.QWebEngineDownloadRequest):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_interrupt_reason(self) -> DownloadInterruptReasonStr:
        """Get reason for download interruption."""
        return DOWNLOAD_INTERRUPT_REASONS.inverse[self.item.interruptReason()]

    def get_state(self) -> DownloadStateStr:
        """Get current state."""
        return DOWNLOAD_STATE.inverse[self.item.state()]

    def set_save_page_format(self, fmt: SavePageFormatStr | Item.SavePageFormat):
        """Set the save page format.

        Args:
            fmt: save page format for the layout
        """
        self.item.setSavePageFormat(SAVE_PAGE_FORMAT.get_enum_value(fmt))

    def get_save_page_format(self) -> SavePageFormatStr:
        """Return current save page format.

        Returns:
            Save page format
        """
        return SAVE_PAGE_FORMAT.inverse[self.item.savePageFormat()]


# if __name__ == "__main__":
#     from prettyqt import widgets

# app = widgets.app()
# item = WebEngineDownloadItem()
# app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from prettyqt import gui
from prettyqt.utils import get_repr


if TYPE_CHECKING:
    from prettyqt.qt import QtCore
    from prettyqt.utils import datatypes


class TextDocumentFragment(gui.QTextDocumentFragment):
    def __repr__(self):
        return get_repr(self, gui.TextDocument(self.toPlainText()))

    def __str__(self):
        return self.toPlainText()

    def __bool__(self):
        return not self.isEmpty()

    @classmethod
    def from_plain_text(cls, text: str) -> Self:
        return cls(cls.fromPlainText(text))

    def write_to_file(
        self,
        path: datatypes.PathType,
        fmt: gui.textdocumentwriter.FormatStr | bytes | QtCore.QByteArray = "plaintext",
    ):
        writer = gui.TextDocumentWriter()
        writer.set_format(fmt)
        writer.set_file_name(path)
        return writer.write(self)


if __name__ == "__main__":
    doc = TextDocumentFragment.from_plain_text("abc")
    print(repr(doc))



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import core
from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


MethodStr = Literal["get", "post"]

METHODS: bidict[MethodStr, QtWebEngineCore.QWebEngineHttpRequest.Method] = bidict(
    get=QtWebEngineCore.QWebEngineHttpRequest.Method.Get,
    post=QtWebEngineCore.QWebEngineHttpRequest.Method.Post,
)


class WebEngineHttpRequest(QtWebEngineCore.QWebEngineHttpRequest):
    def __repr__(self):
        return f"{type(self).__name__}({self.get_url()})"

    def set_headers(self, headers: dict[str, str]):
        for k, v in headers.items():
            self.setHeader(core.QByteArray(k.encode()), core.QByteArray(v.encode()))

    def get_headers(self) -> dict[str, str]:
        return {h.data().decode(): self.header(h).data().decode() for h in self.headers()}

    def set_url(self, url: datatypes.UrlType):
        url = core.Url(url)
        self.setUrl(url)

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def get_post_data(self) -> str:
        return self.postData().data().decode()

    def set_method(
        self, method: MethodStr | QtWebEngineCore.QWebEngineHttpRequest.Method
    ):
        """Set method this WebEngine request is using.

        Args:
            method: method
        """
        self.setMethod(METHODS.get_enum_value(method))

    def get_method(self) -> MethodStr:
        """Get the method this WebEngine request is using.

        Returns:
            method
        """
        return METHODS.inverse[self.method()]


if __name__ == "__main__":
    request = WebEngineHttpRequest()
    request.set_headers(dict(a="b"))
    print(request.get_headers())



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtQml
from prettyqt.utils import bidict, get_repr


PropertyTypeCategoryStr = Literal["invalid", "list", "object", "normal"]

PROPERTY_TYPE_CATEGORY = bidict(
    invalid=QtQml.QQmlProperty.PropertyTypeCategory.InvalidCategory,
    list=QtQml.QQmlProperty.PropertyTypeCategory.List,
    object=QtQml.QQmlProperty.PropertyTypeCategory.Object,
    normal=QtQml.QQmlProperty.PropertyTypeCategory.Normal,
)

TypeStr = Literal["invalid", "property", "signal_property"]

TYPE = bidict(
    invalid=QtQml.QQmlProperty.Type.Invalid,
    property=QtQml.QQmlProperty.Type.Property,
    signal_property=QtQml.QQmlProperty.Type.SignalProperty,
)


class QmlProperty(QtQml.QQmlProperty):
    """Abstracts accessing properties on objects created from QML."""

    def __repr__(self):
        return get_repr(self, self.object())

    def get_method(self) -> core.MetaMethod:
        return core.MetaMethod(self.method())

    def get_property(self) -> core.MetaProperty:
        return core.MetaProperty(self.property())

    def get_property_type_category(self) -> PropertyTypeCategoryStr:
        return PROPERTY_TYPE_CATEGORY.inverse[self.propertyTypeCategory()]

    def get_type(self) -> TypeStr:
        return TYPE.inverse[self.type()]


if __name__ == "__main__":
    prop = QmlProperty()
    method = prop.get_method()
    print(prop)



from __future__ import annotations

import enum

from prettyqt import constants, core, gui, widgets


class AccordionItem(widgets.GroupBox):
    def __init__(self, widget, **kwargs):
        super().__init__(accept_drops=True, context_menu_policy="custom", **kwargs)
        layout = self.set_layout("vertical", spacing=0, margin=6)
        layout.addWidget(widget)
        self._rollout_style = 2
        self._dragdrop_enabled = False
        self.customContextMenuRequested.connect(self.show_menu)
        self._widget = widget
        self._collapsed = False
        self._collapsible = True
        self._clicked = False

    def dragEnterEvent(self, event):
        if not self._dragdrop_enabled:
            return

        source = event.source()
        if (
            source != self
            and source.parent() == self.parent()
            and isinstance(source, AccordionItem)
        ):
            event.acceptProposedAction()

    def get_drag_drop_rect(self) -> core.Rect:
        return core.Rect(self.width() - 25, 7, 10, 6)

    def is_dragdrop_enabled(self) -> bool:
        return self._dragdrop_enabled

    def dragMoveEvent(self, event):
        if not self._dragdrop_enabled:
            return

        source = event.source()
        if (
            source != self
            and source.parent() == self.parent()
            and isinstance(source, AccordionItem)
        ):
            event.acceptProposedAction()

    def dropEvent(self, event):
        widget = event.source()
        layout = self.parent().layout()
        layout.insertWidget(layout.indexOf(self), widget)
        # print(self.parent(), self._widget)
        # self.parent().itemsReordered.emit()

    def expand_collapsed_rect(self) -> core.Rect:
        return core.Rect(0, 0, self.width(), 20)

    # def enterEvent(self, event):
    #     self.accordionWidget().leaveEvent(event)
    #     event.accept()

    # def leaveEvent(self, event):
    #     self.accordionWidget().enterEvent(event)
    #     event.accept()

    def mouseReleaseEvent(self, event):
        if self._clicked and self.expand_collapsed_rect().contains(
            event.position().toPoint()
        ):
            self.toggle_collapsed()
            event.accept()
        else:
            event.ignore()

        self._clicked = False

    def mouseMoveEvent(self, event):
        event.ignore()

    def mousePressEvent(self, event):
        # handle an internal move

        # start a drag event
        if (
            event.button() == constants.MouseButton.LeftButton
            and self.get_drag_drop_rect().contains(event.position().toPoint())
        ):
            # create the pixmap
            pixmap = self.grab()
            # create the mimedata
            mime_data = core.MimeData()
            mime_data.setText(f"ItemTitle::{self.title()}")
            # create the drag
            drag = gui.Drag(self)
            drag.setMimeData(mime_data)
            drag.setPixmap(pixmap)
            drag.setHotSpot(event.position().toPoint())
            if not drag.exec():
                pass
                # self._widget.itemDragFailed.emit(self)
            event.accept()
        elif (
            event.button() == constants.MouseButton.LeftButton
            and self.expand_collapsed_rect().contains(event.position().toPoint())
        ):
            self._clicked = True
            event.accept()

        else:
            event.ignore()

    def isCollapsed(self) -> bool:
        return self._collapsed

    def isCollapsible(self) -> bool:
        return self._collapsible

    def _draw_triangle(self, painter, x, y):
        brush = gui.Brush(
            gui.Color(255, 255, 255, 160), constants.BrushStyle.SolidPattern
        )
        points = (
            (
                core.Point(x + 11, y + 6),
                core.Point(x + 16, y + 11),
                core.Point(x + 11, y + 16),
            )
            if self.isCollapsed()
            else (
                core.Point(x + 9, y + 8),
                core.Point(x + 19, y + 8),
                core.Point(x + 14, y + 13),
            )
        )
        triangle = gui.Polygon(points)
        painter.save()
        painter.setBrush(brush)
        painter.drawPolygon(triangle)
        painter.restore()

    def paintEvent(self, event):
        with gui.Painter(self) as painter:
            painter.setRenderHint(painter.RenderHint.Antialiasing)
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
            x, y, w, h = self.rect().adjusted(0, 0, -1, -1).getRect()
            light_color = self.palette().color(gui.Palette.ColorRole.Light)
            shadow_color = self.palette().color(gui.Palette.ColorRole.Shadow)
            # draw a rounded style
            if self._rollout_style == AccordionWidget.RolloutStyle.Rounded:
                # draw the text
                painter.drawText(
                    x + 33,
                    y + 3,
                    w,
                    16,
                    constants.ALIGN_TOP_LEFT,
                    self.title(),
                )
                # draw the triangle
                self._draw_triangle(painter, x, y)
                # draw the borders
                pen = gui.Pen(light_color)
                pen.setWidthF(0.6)
                painter.setPen(pen)
                r = 8
                painter.drawRoundedRect(x + 1, y + 1, w - 1, h - 1, r, r)
                pen.setColor(shadow_color)
                painter.setPen(pen)
                painter.drawRoundedRect(x, y, w - 1, h - 1, r, r)

            # draw a square style
            if self._rollout_style == AccordionWidget.RolloutStyle.Square:
                painter.drawText(
                    x + 33, y + 3, w, 16, constants.ALIGN_TOP_LEFT, self.title()
                )
                self._draw_triangle(painter, x, y)
                # draw the borders
                pen = gui.Pen(light_color)
                pen.setWidthF(0.6)
                painter.setPen(pen)
                painter.drawRect(x + 1, y + 1, w - 1, h - 1)
                pen.setColor(shadow_color)
                painter.setPen(pen)
                painter.drawRect(x, y, w - 1, h - 1)

            if self._rollout_style == AccordionWidget.RolloutStyle.Maya:
                painter.drawText(
                    x + 33,
                    y + 3,
                    w,
                    16,
                    constants.ALIGN_TOP_LEFT,
                    self.title(),
                )

                painter.setRenderHint(gui.Painter.RenderHint.Antialiasing, False)
                self._draw_triangle(painter, x, y)
                # draw the borders - top
                header_height = 20
                header_rect = core.Rect(x + 1, y + 1, w - 1, header_height)
                header_rect_shadow = core.Rect(x - 1, y - 1, w + 1, header_height + 2)
                # Highlight
                pen = gui.Pen(light_color)
                pen.setWidthF(0.4)
                painter.setPen(pen)
                painter.drawRect(header_rect)
                painter.fillRect(header_rect, gui.Color(255, 255, 255, 18))
                # Shadow
                pen.setColor(self.palette().color(gui.Palette.ColorRole.Dark))
                painter.setPen(pen)
                painter.drawRect(header_rect_shadow)
                if not self.isCollapsed():
                    # draw the lover border
                    pen = gui.Pen(self.palette().color(gui.Palette.ColorRole.Dark))
                    pen.setWidthF(0.8)
                    painter.setPen(pen)
                    offset = header_height + 3
                    body_rect = core.Rect(x, y + offset, w, h - offset)
                    body_rect_shadow = core.Rect(x + 1, y + offset, w + 1, h - offset + 1)
                    painter.drawRect(body_rect)
                    pen.setColor(light_color)
                    pen.setWidthF(0.4)
                    painter.setPen(pen)
                    painter.drawRect(body_rect_shadow)
            # draw a boxed style
            elif self._rollout_style == AccordionWidget.RolloutStyle.Boxed:
                if self.isCollapsed():
                    arect = core.Rect(x + 1, y + 9, w - 1, 4)
                    brect = core.Rect(x, y + 8, w - 1, 4)
                    text = "+"
                else:
                    arect = core.Rect(x + 1, y + 9, w - 1, h - 9)
                    brect = core.Rect(x, y + 8, w - 1, h - 9)
                    text = "-"
                # draw the borders
                pen = gui.Pen(light_color)
                pen.setWidthF(0.6)
                painter.setPen(pen)
                painter.drawRect(arect)
                pen.setColor(shadow_color)
                painter.setPen(pen)
                painter.drawRect(brect)
                painter.setRenderHint(painter.RenderHint.Antialiasing, False)
                color = self.palette().color(gui.Palette.ColorRole.Window).darker(120)
                painter.setBrush(color)
                painter.drawRect(x + 10, y + 1, w - 20, 16)
                painter.drawText(
                    x + 16,
                    y + 1,
                    w - 32,
                    16,
                    constants.ALIGN_CENTER_LEFT,
                    text,
                )
                painter.drawText(
                    x + 10,
                    y + 1,
                    w - 20,
                    16,
                    constants.AlignmentFlag.AlignCenter,
                    self.title(),
                )

            if self.is_dragdrop_enabled():
                rect = self.get_drag_drop_rect()

                # draw the lines
                left = rect.left()
                right = rect.right()
                cy = rect.center().y()

                for y in (cy - 3, cy, cy + 3):
                    painter.drawLine(left, y, right, y)

    def setCollapsed(self, state: bool = True):
        if not self.isCollapsible():
            return
        with self._widget.updates_off():
            self._collapsed = state

            if state:
                self.setMinimumHeight(22)
                self.setMaximumHeight(22)
                self.widget().setVisible(False)
            else:
                self.setMinimumHeight(0)
                self.setMaximumHeight(1000000)
                self.widget().setVisible(True)

            # self._widget.itemCollapsed.emit(self)

    def setCollapsible(self, state: bool = True):
        self._collapsible = state

    def set_dragdrop_enabled(self, mode):
        self._dragdrop_enabled = mode

    def set_rollout_style(self, style):
        self._rollout_style = style

    def show_menu(self):
        if core.Rect(0, 0, self.width(), 20).contains(
            self.mapFromGlobal(gui.Cursor.pos())
        ):
            pass
            # self._widget.itemMenuRequested.emit(self)

    def get_rollout_style(self):
        return self._rollout_style

    def toggle_collapsed(self):
        self.setCollapsed(not self.isCollapsed())

    def widget(self):
        return self._widget


class AccordionWidget(widgets.ScrollArea):
    itemCollapsed = core.Signal(AccordionItem)  # noqa: N815
    itemMenuRequested = core.Signal(AccordionItem)  # noqa: N815
    itemDragFailed = core.Signal(AccordionItem)  # noqa: N815
    itemsReordered = core.Signal()  # noqa: N815

    class RolloutStyle(enum.IntEnum):
        """Rollout style for the widget."""

        Boxed = 1
        Rounded = 2
        Square = 3
        Maya = 4

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            auto_fill_background=False,
            widget_resizable=True,
            mouse_tracking=True,
            frame_shape="no_frame",
            **kwargs,
        )

        widget = widgets.Widget(self)
        # define custom properties
        self._rollout_style = self.RolloutStyle.Rounded
        self._dragdrop_enabled = True
        self._scrolling = False
        self._scroll_init_y = 0
        self._scroll_init_val = 0
        self._ItemClass = AccordionItem
        self._items = {}
        layout = widget.set_layout("vertical", margin=2, spacing=2)
        layout.addStretch(1)
        self.setWidget(widget)

    def setSpacing(self, space_int):
        self.widget().layout().setSpacing(space_int)

    def addItem(
        self,
        title: object,
        widget: object,
        collapsed: object = False,
        index: object = None,
    ) -> object:
        with self.updates_off():
            item = self._ItemClass(widget, title=title, parent=self)
            item.set_rollout_style(self.get_rollout_style())
            item.set_dragdrop_enabled(self.is_dragdrop_enabled())
            layout = self.widget().layout()
            if index is None:  # append if not specified index
                index = layout.count() - 1
            layout.insertWidget(index, item)
            layout.setStretchFactor(item, 0)
            if collapsed:
                item.setCollapsed(collapsed)
            return item

    def clear(self):
        with self.updates_off():
            layout = self.widget().layout()
            while layout.count() > 1:
                item = layout.itemAt(0)
                w = item.widget()
                layout.removeItem(item)
                w.close()
                w.deleteLater()

    # def eventFilter(self, object, event) -> bool:
    #     if event.type() == QtCore.QEvent.Type.MouseButtonPress:
    #         self.mousePressEvent(event)
    #         return True
    #     elif event.type() == QtCore.QEvent.Type.MouseMove:
    #         self.mouseMoveEvent(event)
    #         return True
    #     elif event.type() == QtCore.QEvent.Type.MouseButtonRelease:
    #         self.mouseReleaseEvent(event)
    #         return True
    #     return False

    def count(self):
        return self.widget().layout().count() - 1

    def is_dragdrop_enabled(self) -> bool:
        return self._dragdrop_enabled

    def indexOf(self, widget):
        layout = self.widget().layout()
        return next(
            (
                index
                for index in range(layout.count())
                if layout.itemAt(index).widget().widget() == widget
            ),
            -1,
        )

    def is_boxed_mode(self):
        return self._rollout_style == self.RolloutStyle.Boxed

    def get_rollout_style(self):
        return self._rollout_style

    def item_class(self):
        return self._ItemClass

    def itemAt(self, index):
        layout = self.widget().layout()
        if 0 <= index < layout.count() - 1:
            return layout.itemAt(index).widget()
        return None

    def enterEvent(self, event):
        if self.v_scrollbar.can_scroll():
            widgets.Application.setOverrideCursor(constants.CursorShape.OpenHandCursor)

    def leaveEvent(self, event):
        if self.v_scrollbar.can_scroll():
            widgets.Application.restoreOverrideCursor()

    def mouseMoveEvent(self, event):
        if self._scrolling:
            sbar = self.verticalScrollBar()
            smax = sbar.maximum()
            # calculate the distance moved for the moust point
            dy = event.globalY() - self._scroll_init_y
            # calculate the percentage that is of the scroll bar
            dval = smax * (dy / float(sbar.height()))
            # calculate the new value
            sbar.setValue(self._scroll_init_val - dval)
        event.accept()

    def mousePressEvent(self, event):
        if (
            event.button() == constants.MouseButton.LeftButton
            and self.v_scrollbar.can_scroll()
        ):
            self._scrolling = True
            self._scroll_init_y = event.globalY()
            self._scroll_init_val = self.verticalScrollBar().value()
            widgets.Application.set_override_cursor("closed_hand")
        event.accept()

    def mouseReleaseEvent(self, event):
        if self._scrolling:
            widgets.Application.restoreOverrideCursor()

        self._scrolling = False
        self._scroll_init_y = 0
        self._scroll_init_val = 0
        event.accept()

    def move_item_down(self, index):
        layout = self.widget().layout()
        if (layout.count() - 1) > (index + 1):
            widget = layout.takeAt(index).widget()
            layout.insertWidget(index + 1, widget)

    def move_item_up(self, index):
        if index > 0:
            layout = self.widget().layout()
            widget = layout.takeAt(index).widget()
            layout.insertWidget(index - 1, widget)

    def set_dragdrop_enabled(self, mode: bool):
        self._dragdrop_enabled = mode

        for item in self.findChildren(AccordionItem):
            item.set_dragdrop_enabled(self._dragdrop_enabled)

    def set_item_class(self, item_class):
        self._ItemClass = item_class

    def set_rollout_style(self, rollout_style: RolloutStyle):
        self._rollout_style = rollout_style

        for item in self.findChildren(AccordionItem):
            item.set_rollout_style(self._rollout_style)

    def takeAt(self, index):
        with self.updates_off():
            layout = self.widget().layout()
            widget = None
            if 0 <= index < layout.count() - 1:
                item = layout.itemAt(index)
                widget = item.widget()
                layout.removeItem(item)
                widget.close()
            return widget

    def widgetAt(self, index):
        return item.widget() if (item := self.itemAt(index)) else None

    rollout_style = core.Property(
        int,
        get_rollout_style,
        set_rollout_style,
        doc="Rollout style",
    )
    dragdrop_enabled = core.Property(
        bool,
        is_dragdrop_enabled,
        set_dragdrop_enabled,
        doc="Whether drag&drop is possible.",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = AccordionWidget()
    item = AccordionItem(widgets.LineEdit("test"), title="hallo")
    widget.addItem("tsektjk", AccordionItem(widgets.LineEdit("test"), title="hallo"))
    widget.addItem("tsektjk", AccordionItem(widgets.LineEdit("test"), title="hallo"))
    widget.addItem("tsektjk", widgets.LineEdit("test"))
    widget.addItem("tsektjk", item)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class RenderLinkDelegate(widgets.StyledItemDelegate):
    ID = "render_link"

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        text = index.data()
        if not text:
            return
        option = widgets.StyleOptionViewItem(option)
        self.initStyleOption(option, index)
        painter.save()

        if option.state and widgets.Style.StateFlag.State_MouseOver:
            font = option.font
            font.setUnderline(True)
            painter.setFont(font)
            painter.setPen(option.palette.link().color())
        rect = option.rect.toRectF()
        painter.drawText(rect, constants.ALIGN_CENTER_LEFT, text)
        painter.restore()

    def editorEvent(
        self,
        event,
        model: core.QAbstractItemModel,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        text = index.data()
        font = index.data(constants.FONT_ROLE) or gui.GuiApplication.font()
        # alignment = index.data(constants.ALIGNMENT_ROLE)
        fm = gui.FontMetricsF(font)
        rect = option.rect.toRectF()
        b_rect = fm.get_bounding_rect(rect, constants.ALIGN_CENTER_LEFT, text)
        if (
            event.type() == core.QEvent.Type.MouseButtonPress
            and event.button() == constants.MouseButton.LeftButton  # type: ignore
            and b_rect.contains(event.position())  # type: ignore
        ):
            text = index.data()
            gui.DesktopServices.open_url(text)
            return True
        return False


if __name__ == "__main__":
    app = widgets.app()

    # Create and populate the tableWidget
    table_widget = widgets.TableWidget(4, 2)
    table_widget.set_delegate(RenderLinkDelegate(), column=1)
    table_widget.set_selection_behavior("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Progress"])

    data = [
        ["Mass in B-Minor", "http://www.google.de"],
        ["Three More Foxes", "http://www.google.com"],
        ["Sex Bomb", "http://www.google.de"],
        ["Barbie Girl", "http://www.google.de"],
    ]

    for i, r in enumerate(data):
        table_widget[i, 0] = widgets.TableWidgetItem(r[0])
        item = widgets.TableWidgetItem()
        item.setData(0, r[1])
        item.set_text_alignment("left")
        table_widget[i, 1] = item

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


# COLOR_SPACE = bidict(
#     default=QtGui.QSurfaceFormat.ColorSpace.DefaultColorSpace,
#     srgb=QtGui.QSurfaceFormat.ColorSpace.sRGBColorSpace,
# )

# ColorSpaceStr = Literal["default", "srgb"]

FORMAT_OPTION = bidict(
    stereo_buffers=QtGui.QSurfaceFormat.FormatOption.StereoBuffers,
    debug_context=QtGui.QSurfaceFormat.FormatOption.DebugContext,
    deprecated_functions=QtGui.QSurfaceFormat.FormatOption.DeprecatedFunctions,
    reset_notification=QtGui.QSurfaceFormat.FormatOption.ResetNotification,
    protected_content=QtGui.QSurfaceFormat.FormatOption.ProtectedContent,
)

FormatOptionStr = Literal[
    "stereo_buffers",
    "debug_context",
    "deprecated_functions",
    "reset_notification",
    "protected_content",
]

OPEN_GL_CONTEXT_PROFILE = bidict(
    none=QtGui.QSurfaceFormat.OpenGLContextProfile.NoProfile,
    core=QtGui.QSurfaceFormat.OpenGLContextProfile.CoreProfile,
    compatibility=QtGui.QSurfaceFormat.OpenGLContextProfile.CompatibilityProfile,
)

OpenGLContextProfileStr = Literal["none", "core", "compatibility"]

RENDERABLE_TYPE = bidict(
    default=QtGui.QSurfaceFormat.RenderableType.DefaultRenderableType,
    open_gl=QtGui.QSurfaceFormat.RenderableType.OpenGL,
    open_gl_es=QtGui.QSurfaceFormat.RenderableType.OpenGLES,
    open_vg=QtGui.QSurfaceFormat.RenderableType.OpenVG,
)

RenderableTypeStr = Literal["default", "open_gl", "open_gl_es", "open_vg"]

SWAP_BEHAVIOR = bidict(
    default=QtGui.QSurfaceFormat.SwapBehavior.DefaultSwapBehavior,
    single=QtGui.QSurfaceFormat.SwapBehavior.SingleBuffer,
    double=QtGui.QSurfaceFormat.SwapBehavior.DoubleBuffer,
    triple=QtGui.QSurfaceFormat.SwapBehavior.TripleBuffer,
)

SwapBehaviorStr = Literal["default", "single", "double", "triple"]


class SurfaceFormatMixin:
    def __repr__(self):
        return get_repr(self)

    def get_profile(self) -> OpenGLContextProfileStr:
        """Get the current OpenGl profile.

        Returns:
            OpenGL profile
        """
        return OPEN_GL_CONTEXT_PROFILE.inverse[self.profile()]

    def get_renderable_type(self) -> RenderableTypeStr:
        """Get the current renderable type.

        Returns:
            renderable type
        """
        return RENDERABLE_TYPE.inverse[self.renderableType()]

    def get_swap_behavior(self) -> SwapBehaviorStr:
        """Get the current swap behavior.

        Returns:
            swap behavior
        """
        return SWAP_BEHAVIOR.inverse[self.swapBehavior()]


class SurfaceFormat(SurfaceFormatMixin, QtGui.QSurfaceFormat):
    pass



from __future__ import annotations

import enum
import io
import logging
from typing import ClassVar
import xml.etree.ElementTree as ET

from prettyqt import constants, core, itemmodels
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class XmlColumnItem(itemmodels.ColumnItem):
    def get_data(self, item: XmlModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case XmlModel.Roles.NodeRole:
                return item.obj


class TagColumn(XmlColumnItem):
    name = "Tag"
    doc = "Tag"

    def get_data(self, item: XmlModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.tag
        return super().get_data(item, role)


class TextColumn(XmlColumnItem):
    name = "Text"
    doc = "Text"

    def get_data(self, item: XmlModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.text
        return super().get_data(item, role)


class TailColumn(XmlColumnItem):
    name = "Tail"
    doc = "Tail"

    def get_data(self, item: XmlModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.tail
        return super().get_data(item, role)


class AttributeColumn(XmlColumnItem):
    name = "Attribute"
    doc = "Attribute"

    def get_data(self, item: XmlModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.attrib
        return super().get_data(item, role)


class BaseXmlModel(itemmodels.ColumnItemModel):
    ICON = "mdi.xml"
    COLUMNS: ClassVar = [TagColumn, TextColumn, TailColumn, AttributeColumn]

    def __init__(self, obj, **kwargs):
        super().__init__(obj=obj, columns=self.COLUMNS, show_root=True, **kwargs)

    class Roles(enum.IntEnum):
        NodeRole = constants.USER_ROLE + 24245

    def _has_children(self, item: XmlModel.TreeItem) -> bool:
        return len(item.obj) > 0

    def _fetch_object_children(self, item: XmlModel.TreeItem) -> list[XmlModel.TreeItem]:
        return [self.TreeItem(obj=i) for i in item.obj]


class XmlModel(BaseXmlModel):
    """Semi-lazy xml model. Fetches all direct child nodes when needed.

    Model cant be modified, that only really makes sense for a full DOM implementation.
    """

    SUPPORTS = io.StringIO | str | datatypes.IsTreeIterator | ET.ElementTree

    def __init__(
        self,
        obj: str | datatypes.IsTreeIterator | ET.ElementTree,
        show_root: bool = True,
        **kwargs,
    ):
        match obj:
            case io.StringIO():
                context = ET.iterparse(obj, events=("start",))
                _, root = next(context)
            case str():
                context = ET.iterparse(io.StringIO(obj), events=("start",))
                _, root = next(context)
            case datatypes.IsTreeIterator():
                _, root = next(obj)
            case ET.ElementTree():
                xml_str = ET.tostring(obj._root, encoding="unicode")
                context = ET.iterparse(io.StringIO(xml_str), events=("start",))
                _, root = next(context)
            case _:
                raise TypeError(obj)
        super().__init__(obj=root, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case datatypes.IsTreeIterator() | ET.ElementTree():
                return True
            case _:
                return False

    def get_parent_node(self, node_or_index: core.ModelIndex | ET.Element):
        # only lxml has Element.getparent()
        # for builtin Elements we need to go through indexes since we do not have a
        # full ElementTree.
        # might be worth having two XML models, a lazy one and a full-featured one.
        # Then we could use xpath to get parent here.
        if isinstance(node_or_index, core.ModelIndex):
            index = node_or_index
        elif indexes := self.search_tree(
            node_or_index, role=self.Roles.NodeRole, max_results=1
        ):
            index = indexes[0]
        else:
            return None
        parent = index.parent()
        return parent.data(self.Roles.NodeRole)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    xml = """<root>
<element key='value'>text</element>
<element><sub>text</sub></element>
<empty-element xmlns="http://testns/" />
</root>
"""
    xml = ET.parse(io.StringIO(xml))
    table = widgets.TreeView()
    model = XmlModel(xml, parent=table)
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.set_model(model)

    def test(new, old):
        node = new.data(new.model().Roles.NodeRole)
        new.model().get_parent_node(node)

    table.selectionModel().currentChanged.connect(test)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import ast

from prettyqt import gui


class PythonCodeValidator(gui.Validator):
    """Validator which checks whether given string is valid Python code."""

    ID = "python_code"

    def __eq__(self, other: object):
        return isinstance(other, PythonCodeValidator)

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        try:
            ast.parse(text)
        except SyntaxError:
            return self.State.Intermediate, text, pos
        else:
            return self.State.Acceptable, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = PythonCodeValidator()
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core


if TYPE_CHECKING:
    from prettyqt.qt import QtHelp


class HelpLink:
    """Struct provides the data associated with a help link."""

    def __init__(self, item: QtHelp.QHelpLink):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_url(self) -> core.Url:
        return core.Url(self.url)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import widgets
from prettyqt.utils import colors


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class GraphicsColorizeEffect(
    widgets.GraphicsEffectMixin, widgets.QGraphicsColorizeEffect
):
    def set_color(self, color: datatypes.ColorType):
        color = colors.get_color(color)
        super().setColor(color)


if __name__ == "__main__":
    app = widgets.app()
    effect = GraphicsColorizeEffect()
    effect.set_color("window_role")
    print(effect.color())



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from prettyqt import core
from prettyqt.qt import QtWebEngineCore


if TYPE_CHECKING:
    from collections.abc import Callable


class WebEngineUrlSchemeHandler(
    core.ObjectMixin, QtWebEngineCore.QWebEngineUrlSchemeHandler
):
    pass


class CallbackWebEngineUrlSchemeHandler(WebEngineUrlSchemeHandler):
    def __init__(
        self,
        callback: Callable[[QtWebEngineCore.QWebEngineUrlRequestJob], Any],
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.callback = callback

    def requestStarted(self, request: QtWebEngineCore.QWebEngineUrlRequestJob):
        self.callback(request)


if __name__ == "__main__":
    item = WebEngineUrlSchemeHandler()



"""Provides QtCharts classes and functions."""

from prettyqt.qt import PYQT6, PYSIDE6


if PYQT6:
    try:
        from PyQt6.QtCharts import *
    except ImportError:
        raise ModuleNotFoundError(
            "The QtChart module was not found. "
            "It needs to be installed separately for PyQt6."
        )

elif PYSIDE6:
    from PySide6.QtCharts import *
else:
    raise ModuleNotFoundError("No Qt bindings could be found")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, TypeVar

from prettyqt import constants, core
from prettyqt.utils import datatypes, helpers


if TYPE_CHECKING:
    from collections.abc import Callable


T = TypeVar("T", bound=core.QObject)

logger = logging.getLogger(__name__)

# TYPES = {
#     bool: 1,
#     int: 2,
#     str: 10,
#     float: 38,
#     QtGui.QColor: 67,
#     QtGui.QCursor: 74,
#     core.QDate: 14,
#     core.QSize: 21,
#     core.QTime: 15,
#     list: 9,
#     QtGui.QPolygon: 71,
#     QtGui.QPolygonF: 86,
#     QtGui.QColor: 67,
#     QtGui.QColorSpace: 87,
#     core.QSizeF: 22,
#     core.QRectF: 20,
#     core.QLine: 23,
#     QtGui.QTextLength: 77,
#     dict: 8,
#     QtGui.QIcon: 69,
#     QtGui.QPen: 76,
#     core.QLineF: 24,
#     QtGui.QTextFormat: 78,
#     core.QRect: 19,
#     core.QPoint: 25,
#     core.QUrl: 17,
#     core.QRegularExpression: 44,
#     core.QDateTime: 16,
#     core.QPointF: 26,
#     QtGui.QPalette: 68,
#     QtGui.QFont: 64,
#     QtGui.QBrush: 66,
#     QtGui.QRegion: 72,
#     QtGui.QImage: 70,
#     QtGui.QKeySequence: 75,
#     QtWidgets.QSizePolicy: 121,
#     QtGui.QPixmap: 65,
#     core.QLocale: 18,
#     QtGui.QBitmap: 73,
#     QtGui.QMatrix4x4: 81,
#     QtGui.QVector2D: 82,
#     QtGui.QVector3D: 83,
#     QtGui.QVector4D: 84,
#     QtGui.QQuaternion: 85,
#     core.QEasingCurve: 29,
#     core.QJsonValue: 45,
#     core.QJsonDocument: 48,
#     core.QModelIndex: 42,
#     core.QPersistentModelIndex: 50,
#     core.QUuid: 30,
#     "user": 1024,
# }


class MetaObject:
    """Contains meta-information about Qt objects."""

    def __init__(self, metaobject: core.QMetaObject):
        self.item = metaobject

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_super_class(self) -> MetaObject | None:
        """Get SuperClass MetaObject."""
        return MetaObject(klass) if (klass := self.superClass()) is not None else None

    def get_all_super_classes(self) -> list[MetaObject]:
        """Get SuperClass MetaObject."""
        klasses = []
        while klass := self.superClass():
            klasses.append(MetaObject(klass))
        return klasses

    def get_name(self) -> str:
        """Get MetaObject class name."""
        return self.className()

    def get_class_info(self, include_super: bool = True) -> dict[str, str]:
        """Get MetaObject class info.

        Arguments:
            include_super: Whether to include class info from parent classes.
        """
        start = 0 if include_super else self.item.classInfoOffset()
        count = self.item.classInfoCount()
        classinfos = [self.item.classInfo(i) for i in range(start, count)]
        return {i.name(): i.value() for i in classinfos}

    def get_method(self, index: int | str) -> core.MetaMethod:
        """Get MetaMethod based on index or name.

        Arguments:
            index: index or method name
        """
        if isinstance(index, int):
            method = core.MetaMethod(self.item.method(index))
            if not method.isValid():
                raise KeyError(index)
            return method
        for method in self.get_methods():
            if method.get_name() in [index, helpers.to_lower_camel(index)]:
                return method
        raise KeyError(index)

    def get_enum(self, index: int | str) -> core.MetaEnum:
        """Get MetaEnum based on index or name.

        Arguments:
            index: index or Enum name
        """
        if isinstance(index, int):
            enum = core.MetaEnum(self.item.enumerator(index))
            if not enum.isValid():
                raise KeyError(index)
            return enum
        for enumerator in self.get_enums():
            if enumerator.get_name() in [index, helpers.to_lower_camel(index)]:
                return enumerator
        raise KeyError(index)

    def has_property(self, name: str):
        """Check if a property with given name exists.

        Only checks for non-dynamic properties.

        Arguments:
            name: property name
        """
        try:
            self.get_property(name)
        except KeyError:
            return False
        else:
            return True

    def get_property(self, index: int | str) -> core.MetaProperty:
        """Get MetaProperty based on index or name.

        Only returns non-dynamic properties.

        Arguments:
            index: index or property name
        """
        if isinstance(index, int):
            prop = core.MetaProperty(self.item.property(index))
            if not prop.isValid():
                raise KeyError(index)
            return prop
        for prop in self.get_properties():
            if prop.get_name() in [index, helpers.to_lower_camel(index)]:
                return prop
        raise KeyError(index)

    def get_constructor(self, index: int | str) -> core.MetaMethod:
        """Get ctor MetaMethod based on index or name.

        Arguments:
            index: index or constructor name
        """
        if isinstance(index, int):
            method = core.MetaMethod(self.item.constructor(index))
            if not method.isValid():
                raise KeyError(index)
            return method
        for method in self.get_constructors():
            if method.get_name() in [index, helpers.to_lower_camel(index)]:
                return method
        raise KeyError(index)

    def get_methods(
        self,
        include_super: bool = True,
        type_filter: core.metamethod.MethodTypeStr | None = None,
        filter_internal: bool = True,
    ) -> list[core.MetaMethod]:
        """Get all MetaMethods based on given criteria.

        Arguments:
            include_super: Whether to include Methods from parent classes
            type_filter: Method type to filter for.
            filter_internal: Filter Qt-internal methods
        """
        start = 0 if include_super else self.item.methodOffset()
        methods = [
            method
            for i in range(start, self.item.methodCount())
            if not (method := self.get_method(i)).get_name().startswith("_q_")
            or not filter_internal
        ]
        if type_filter is None:
            return methods
        return [i for i in methods if i.get_method_type() == type_filter]

    def get_enums(self, include_super: bool = True) -> list[core.MetaEnum]:
        """Get all MetaEnums based on given criteria.

        Arguments:
            include_super: Whether to include Enums from parent classes.
        """
        start = 0 if include_super else self.item.enumeratorOffset()
        return [self.get_enum(i) for i in range(start, self.item.enumeratorCount())]

    def get_constructors(self) -> list[core.MetaMethod]:
        """Get all ctor MetaMethods."""
        count = self.item.constructorCount()
        return [core.MetaMethod(self.item.constructor(i)) for i in range(count)]

    def get_properties(
        self,
        include_super: bool = True,
        only_writable: bool = False,
        only_stored: bool = False,
        only_bindable: bool = False,
        only_designable: bool = False,
        only_final: bool = False,
        only_required: bool = False,
        only_enum_type: bool = False,
        only_flag_type: bool = False,
        only_with_notifiers: bool = False,
        only_with_type_name: str = "",
    ) -> list[core.MetaProperty]:
        """Get all MetaProperties based on given criteria.

        Arguments:
            include_super: Whether to include properties from parent classes,
            only_writable: Whether to filter for writable properties.
            only_stored: Whether to filter for stored properties.
            only_bindable: Whether to filter for bindable properties.
            only_designable: Whether to filter for designable properties.
            only_final: Whether to filter for final properties.
            only_required: Whether to filter for required properties.
            only_enum_type: Whether to filter for Enum type properties.
            only_flag_type: Whether to filter for Flag_type properties.
            only_with_notifiers: Whether to filter for properties with notifier.
            only_with_type_name: Only include properties with given give name as type.
        """
        start = 0 if include_super else self.item.propertyOffset()
        count = self.item.propertyCount()
        prop_list = []
        for i in range(start, count):
            prop = self.item.property(i)
            if (
                (only_writable and not prop.isWritable())
                or (only_stored and not prop.isStored())
                or (only_bindable and not prop.isBindable())
                or (only_designable and not prop.isDesignable())
                or (only_final and not prop.isFinal())
                or (only_required and not prop.isRequired())
                or (only_enum_type and not prop.isEnumType())
                or (only_flag_type and not prop.isFlagType())
                or (only_with_notifiers and not prop.hasNotifier())
                or (only_with_type_name and prop.typeName() != only_with_type_name)
            ):
                continue
            prop_list.append(core.MetaProperty(prop))
        return prop_list

    def get_property_values(
        self, qobject: core.QObject, cast_types: bool = False
    ) -> dict[str, Any]:
        """Get a dictionary containing all MetaProperties values from given qobject.

        Arguments:
            qobject: QObject to get properties from
            cast_types: Whether to cast types to PrettyQt classes.
        """
        vals = {prop.get_name(): prop.read(qobject) for prop in self.get_properties()}
        if cast_types:
            return {k: datatypes.make_serializable(v) for k, v in vals.items()}
        return vals

    def get_signals(
        self, include_super: bool = True, only_notifiers: bool = False
    ) -> list[core.MetaMethod]:
        """Get all signal MetaMethods based on given criteria.

        Arguments:
            include_super: Whether to include Signals from parent classes
            only_notifiers: Whether to filter for property notifier signals
        """
        if only_notifiers:
            return [  # type: ignore
                prop.get_notify_signal()
                for prop in self.get_properties(include_super)
                if prop.hasNotifySignal()
            ]
        return self.get_methods(include_super=include_super, type_filter="signal")

    def get_slots(self, include_super: bool = True) -> list[core.MetaMethod]:
        """Get all slot MetaMethods based on given criteria.

        Arguments:
            include_super: Whether to include Slots from parent classes
        """
        return self.get_methods(include_super=include_super, type_filter="slot")

    def get_plain_methods(self, include_super: bool = True) -> list[core.MetaMethod]:
        """Get all plain MetaMethods based on given criteria.

        Arguments:
            include_super: Whether to include plain methods from parent classes
        """
        return self.get_methods(include_super=include_super, type_filter="method")

    def get_meta_type(self) -> core.MetaType:
        """Get Meta type of this MetaObject."""
        return core.MetaType(self.metaType().id())

    def get_user_property(self) -> core.MetaProperty | None:
        """Get MetaProperty marked as userprop."""
        return core.MetaProperty(p) if (p := self.userProperty()).isValid() else None

    # just experimenting
    @classmethod
    def invoke_method(
        cls,
        obj: core.QObject,
        method: str,
        *args,
        connection_type: constants.ConnectionTypeStr = "auto",
    ):
        conn = constants.CONNECTION_TYPE[connection_type]
        args = tuple(core.Q_ARG(type(arg), arg) for arg in args)
        return cls.invokeMethod(obj, method, conn, *args)

    def get_new_instance(self, *args, **kwargs):
        args = tuple(core.Q_ARG(type(i), i) for i in args)
        kwargs = {k: core.Q_ARG(type(v), v) for k, v in kwargs.items()}
        # requires core.QGenericArgumentHolder for PySide6
        self.newInstance(*args, **kwargs)

    def connect_signals(
        self,
        source_qobject: core.QObject,
        fn_or_qobject: Callable | core.QObject,
        only_notifiers: bool = False,
    ) -> list[core.QMetaObject.Connection]:
        """Connect all signals of a given qobject.

        Either connect all signals to a function or connect each signal
        to the corresponding signal of the receiver.
        """
        handles = []
        for signal in self.get_signals(only_notifiers=only_notifiers):
            signal_name = signal.get_name()
            if not hasattr(source_qobject, signal_name):
                # PyQt6 reports applicationNameChanged for QCoreApplication,
                # but it doesnt exist...
                logger.warning("Signal %r does not exist.", signal_name)
                continue
            signal_instance = getattr(source_qobject, signal_name)
            slot = (
                getattr(fn_or_qobject, signal_name)
                if isinstance(fn_or_qobject, core.QObject)
                else fn_or_qobject
            )
            handle = signal_instance.connect(slot)
            handles.append(handle)
        logger.debug("connected %s signals to %s.", len(handles), fn_or_qobject)
        return handles

    def copy(self, qobject: T, forward_signals: bool = True) -> T:
        """Create a copy of given QObject.

        Arguments:
            qobject: QObject to create a copied instance for.
            forward_signals: Whether to connect all signals from qobject to the new
                             instance.
        """
        try:
            new = type(qobject)()
        except TypeError:
            # this should should cover most cases.
            new = type(qobject)(qobject.orientation())
        for prop in self.get_properties(only_writable=True):
            val = prop.read(qobject)
            prop.write(new, val)
        if forward_signals:
            self.connect_signals(new, qobject)
        logger.debug("copied %r", qobject)
        return new

    @classmethod
    def copy_properties_to(cls, source: core.QObject, target: core.QObject):
        """Sets all properties of target to value of source.

        Only sets properties which exist for both QObjects.

        Arguments:
            source: Source QObject
            target: Target QObject
        """
        source_metaobj = cls(source.metaObject())
        target_metaobj = cls(target.metaObject())
        for prop in target_metaobj.get_properties(only_writable=True):
            if source_metaobj.has_property(prop_name := prop.get_name()):
                target.setProperty(prop_name, source.property(prop_name))

    def get_property_class_affiliations(self) -> dict[str, list[core.MetaProperty]]:
        """Get a mapping of class -> property affiliations."""
        mapper = {}
        metaclass = self
        while metaclass is not None:
            mapper[metaclass.get_name()] = [
                metaclass.get_property(i)
                for i in range(metaclass.propertyOffset(), metaclass.propertyCount())
            ]
            metaclass = metaclass.get_super_class()
        return mapper


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    h = widgets.HeaderView("horizontal")
    metaobj = h.get_metaobject()
    kls = metaobj.get_property_class_affiliations()
    # print(kls)



from __future__ import annotations

from prettyqt import network
from prettyqt.qt import QtCore


class NetworkDatagram(network.QNetworkDatagram):
    """The data and metadata of a UDP datagram."""

    def get_destination_address(self) -> network.HostAddress:
        return network.HostAddress(self.destinationAddress())

    def get_sender_address(self) -> network.HostAddress:
        return network.HostAddress(self.senderAddress())

    def set_data(self, data: str):
        self.setData(QtCore.QByteArray(data.encode()))

    def get_data(self) -> str:
        return self.data().data().decode()


if __name__ == "__main__":
    datagram = NetworkDatagram()



from __future__ import annotations

from prettyqt import gui


class RasterWindow(gui.PaintDeviceWindowMixin, gui.QRasterWindow):
    """Convenience class for using QPainter on a QWindow."""


if __name__ == "__main__":
    app = gui.app()



from __future__ import annotations

from prettyqt import constants, widgets
from prettyqt.utils import datatypes


class GraphicsRotation(widgets.GraphicsTransformMixin, widgets.QGraphicsRotation):
    """Rotation transformation around a given axis."""

    def set_axis(self, axis: constants.AxisStr | constants.Axis):
        self.setAxis(constants.AXIS.get_enum_value(axis))

    def set_origin(self, origin: datatypes.Vector3DType):
        self.setOrigin(datatypes.to_vector3d(origin))


if __name__ == "__main__":
    rotation = GraphicsRotation()
    rotation.set_origin((1, 1, 1))



from __future__ import annotations

import logging

from prettyqt import constants, core
from prettyqt.qt import QtWidgets


logger = logging.getLogger(__name__)


class AbstractItemDelegateMixin(core.ObjectMixin):
    # this pattern doesnt work for PySide6 yet (cant pass kwargs for class definition)
    # _registry = {}

    # def __init_subclass__(cls, identifier=None, **kwargs):
    #     super().__init_subclass__(**kwargs)
    #     if identifier is not None:
    #         logger.debug(f"registering delegate {cls} {identifier!r}")
    #         if (
    #             identifier in cls._registry
    #             and cls._registry[identifier].__name__ != cls.__name__
    #         ):
    #             raise NameError(f"Delegate with id {identifier!r} already registered.")
    #         cls._registry[identifier] = cls

    @staticmethod
    def _data_for_index(
        index: core.ModelIndex, role: constants.ItemDataRole = constants.USER_ROLE
    ):
        # using index.data() sometimes casts stuff in PyQt6
        model = index.model()
        return model.data(index, role)


class AbstractItemDelegate(AbstractItemDelegateMixin, QtWidgets.QAbstractItemDelegate):
    """Used to display and edit data items from a model."""


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui


if TYPE_CHECKING:
    from prettyqt.qt import QtCore


class BlacklistValidator(gui.Validator):
    """Validator which excludes a fixed list of strings."""

    ID = "blacklist"

    def __init__(
        self,
        options: list[str] | None = None,
        parent: QtCore.QObject | None = None,
    ):
        super().__init__(parent)
        self._options = options or []

    def __eq__(self, other: object):
        return isinstance(other, BlacklistValidator) and self._options == other._options

    def validate(self, text: str, pos: int = 0):
        if any(option == text for option in self._options):
            return self.State.Invalid, text, pos
        return self.State.Acceptable, text, pos


class NotZeroValidator(BlacklistValidator):
    ID = "not_zero"

    def __init__(self, parent: QtCore.QObject | None = None):
        super().__init__(parent=parent)
        self._options = ["0"]


class NotEmptyValidator(BlacklistValidator):
    ID = "not_empty"

    def __init__(self, parent: QtCore.QObject | None = None):
        super().__init__(parent=parent)
        self._options = [""]


if __name__ == "__main__":
    from prettyqt import widgets

    val = BlacklistValidator(["a", "abc", "xyz"])
    app = widgets.app()
    widget = widgets.LineEdit("Thisisatest")
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

import pathlib
from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


IteratorFlagStr = Literal[
    "none",
    "subdirectories",
    "follow_symlinks",
]

ITERATOR_FLAG: bidict[IteratorFlagStr, core.QDirIterator.IteratorFlag] = bidict(
    none=core.QDirIterator.IteratorFlag.NoIteratorFlags,
    subdirectories=core.QDirIterator.IteratorFlag.Subdirectories,
    follow_symlinks=core.QDirIterator.IteratorFlag.FollowSymlinks,
)


class DirIterator(core.QDirIterator):
    """Iterator for directory entrylists."""

    def __iter__(self):
        return self

    def __next__(self):
        if self.hasNext():
            return self.next()
        raise StopIteration

    def get_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.filePath())

    def get_file_info(self) -> core.FileInfo:
        return core.FileInfo(self.fileInfo())


if __name__ == "__main__":
    it = core.DirIterator("C:/Intel")



from __future__ import annotations

from collections.abc import Iterable, Mapping
import functools
from typing import TYPE_CHECKING, Any

from prettyqt import constants, core, iconprovider, widgets
from prettyqt.utils import listdelegators


if TYPE_CHECKING:
    from collections.abc import Iterator

    from prettyqt.qt import QtGui
    from prettyqt.utils import datatypes


class NoData:
    pass


class ListWidget(widgets.ListViewMixin, widgets.QListWidget):
    """Item-based list widget."""

    value_changed = core.Signal(object)

    def __init__(self, *args, selection_mode: str = "single", **kwargs):
        super().__init__(*args, selection_mode=selection_mode, **kwargs)
        self.itemSelectionChanged.connect(self.on_index_change)

    def __repr__(self):
        return f"{type(self).__name__}: {self.count()} items"

    def __getitem__(
        self, row: int | slice
    ) -> widgets.QListWidgetItem | listdelegators.ListDelegator[widgets.QListWidgetItem]:
        match row:
            case int():
                item = self.item(row)
                if item is None:
                    raise KeyError(row)
                return item
            case slice():
                count = self.itemCount() if row.stop is None else row.stop
                values = list(range(count)[row])
                ls = [self.item(i) for i in values]
                return listdelegators.ListDelegator(ls)
            case _:
                raise TypeError(row)

    def __delitem__(self, row: int):
        self.takeItem(row)

    def __add__(self, other: widgets.QListWidgetItem):
        self.addItem(other)
        return self

    def __iter__(self) -> Iterator[widgets.QListWidgetItem]:
        return iter(self.get_children())

    def __len__(self) -> int:
        return self.count()

    def sort(self, reverse: bool = False):
        order = constants.DESCENDING if reverse else constants.ASCENDING
        self.sortItems(order)

    def on_index_change(self):
        data = self.get_value()
        self.value_changed.emit(data)

    def get_children(self) -> listdelegators.ListDelegator[widgets.QListWidgetItem]:
        items = [self.item(row) for row in range(self.count())]
        return listdelegators.ListDelegator(items)

    def add_items(self, items: Iterable | Mapping):
        match items:
            case Mapping():
                for k, v in items.items():
                    self.add(v, k)
            case Iterable():
                for i in items:
                    if isinstance(i, tuple | list):
                        self.add(*i)
                    else:
                        self.add(i)
            case _:
                raise TypeError(items)

    @functools.singledispatchmethod
    def add_item(
        self,
        name: str = "",
        icon: datatypes.IconType = None,
        data: dict | None = None,
        foreground: QtGui.QBrush | None = None,
        background: QtGui.QBrush | None = None,
        font: QtGui.QFont | None = None,
        selected: bool | None = None,
        status_tip: str = "",
        tool_tip: str = "",
        whats_this: str | None = None,
        # text_alignment: Optional[str] = None,
        checkstate: constants.StateStr | constants.CheckState | None = None,
        flags: constants.ItemFlag | None = None,
        size_hint: datatypes.SizeType | None = None,
        is_user_type: bool = False,
    ) -> widgets.ListWidgetItem:
        typ = 1 if is_user_type else 0
        item = widgets.ListWidgetItem(name, self, typ)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.set_data(k, v)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if selected:
            item.setSelected(selected)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.set_size_hint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.addItem(item)
        return item

    @add_item.register
    def _(self, listitem: widgets.QListWidgetItem):
        self.addItem(listitem)

    def add(self, label: str, data=NoData, icon: datatypes.IconType = None):
        if data is NoData:
            data = label
        item = widgets.ListWidgetItem(label)
        item.set_icon(icon)
        item.setData(constants.USER_ROLE, data)  # type: ignore
        self.addItem(item)

    def get_value(self) -> list[Any]:
        return [i.data(constants.USER_ROLE) for i in self.selectedItems()]  # type: ignore

    def set_value(self, value):
        for i in self.get_children():
            if i.data(constants.USER_ROLE) in value:  # type: ignore
                self.setCurrentItem(i)
                break

    def scroll_to_item(
        self,
        item: widgets.QListWidgetItem,
        mode: (
            widgets.abstractitemview.ScrollHintStr | widgets.QAbstractItemView.ScrollHint
        ) = "ensure_visible",
    ):
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT.get_enum_value(mode))

    def find_items(
        self,
        text: str,
        column: int = 0,
        mode: constants.MatchFlagStr | constants.MatchFlag = "exact",
        recursive: bool = False,
        case_sensitive: bool = False,
    ) -> listdelegators.ListDelegator[widgets.QListWidgetItem]:
        flag = constants.MATCH_FLAGS.get_enum_value(mode)
        if recursive:
            flag |= constants.MatchFlag.MatchRecursive
        if case_sensitive:
            flag |= constants.MatchFlag.MatchCaseSensitive
        items = self.findItems(text, flag, column)  # type: ignore
        return listdelegators.ListDelegator(items)


if __name__ == "__main__":
    app = widgets.app()
    widget = ListWidget()
    for i in range(100):
        widget.add_item(str(i))
    widget.add("test", icon="mdi.timer")
    widget.add("test", icon="mdi.timer")
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, core, gui, itemmodels
from prettyqt.utils import colors


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class SliceAppearanceProxyModel(itemmodels.SliceIdentityProxyModel):
    ID = "slice_appearance"
    ICON = "mdi.palette-outline"

    def __init__(
        self,
        foreground: gui.QColor | gui.QBrush | str | None = None,
        background: gui.QColor | gui.QBrush | str | None = None,
        font: str | gui.QFont | None = None,
        alignment: constants.AlignmentFlag | constants.AlignmentStr | None = None,
        override: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._foreground = gui.QColor()
        self.set_foreground(foreground)
        self._background = gui.QColor()
        self.set_background(background)
        self._font = gui.QFont()
        self.set_font(font)
        self._alignment = constants.ALIGN_CENTER_LEFT
        self.set_alignment(alignment)
        self._override = override

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        data = super().data(index, role)
        if not self.indexer_contains(index):
            return data
        if self._override or data is None:
            match role:
                case constants.FOREGROUND_ROLE:
                    return self._foreground if self._foreground.isValid() else None
                case constants.BACKGROUND_ROLE:
                    return self._background if self._background.isValid() else None
                case constants.FONT_ROLE:
                    return self._font
                case constants.ALIGNMENT_ROLE:
                    return self._alignment
        return super().data(index, role)

    def set_font(self, font: gui.QFont | str | None):
        self._font = gui.QFont(font) if font else gui.QFont()
        self.update_all()

    def get_font(self) -> gui.QFont:
        return self._font or gui.QFont()

    def set_foreground(self, foreground: datatypes.ColorAndBrushType | None):
        match foreground:
            case None:
                foreground = gui.QColor()
            case gui.QBrush():
                foreground = foreground.color()
            case _:
                foreground = colors.get_color(foreground).as_qt()
        self._foreground = foreground
        self.update_all()

    def get_foreground(self) -> gui.QColor:
        return self._foreground

    def set_background(self, background: datatypes.ColorAndBrushType | None):
        match background:
            case None:
                background = gui.QColor()
            case gui.QBrush():
                background = background.color()
            case _:
                background = colors.get_color(background).as_qt()
        self._background = background
        self.update_all()

    def get_background(self) -> gui.QColor:
        return self._background

    def set_alignment(
        self, alignment: constants.AlignmentFlag | constants.AlignmentStr | None
    ):
        match alignment:
            case None:
                alignment = constants.ALIGN_CENTER_LEFT
            case str() | constants.AlignmentFlag():
                alignment = constants.ALIGNMENTS.get_enum_value(alignment)
        self._alignment = alignment
        self.update_all()

    def get_alignment(self) -> constants.AlignmentFlag:
        return self._alignment

    font_value = core.Property(
        gui.QFont,
        get_font,
        set_font,
        doc="Font to use for overriding",
    )
    foreground_value = core.Property(
        gui.QColor,
        get_foreground,
        set_foreground,
        doc="Foreground to use for overriding",
    )
    background_value = core.Property(
        gui.QColor,
        get_background,
        set_background,
        doc="Background to use for overriding",
    )
    alignment_value = core.Property(
        constants.AlignmentFlag,
        get_alignment,
        set_alignment,
        doc="Alignment to use for overriding",
    )


if __name__ == "__main__":
    from prettyqt import debugging, gui, widgets

    app = widgets.app()
    table = debugging.example_table()
    table.proxifier[:, 1].style()
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import gui
from prettyqt.utils import get_repr


MAX_VAL = 1_000_000_000  # could be 2_147_483_647, but no point in random number


class IntValidator(gui.ValidatorMixin, gui.QIntValidator):
    """Validator that ensures a string contains a integer within a specified range."""

    ID = "integer_classic"

    def __repr__(self):
        return get_repr(self, self.bottom(), self.top())

    def __getstate__(self):
        return dict(bottom=self.bottom(), top=self.top())

    def __reduce__(self):
        return type(self), (self.bottom(), self.top()), None

    def __eq__(self, other: object):
        return (
            self.bottom() == other.bottom() and self.top() == other.top()
            if isinstance(other, type(self))
            else False
        )

    def set_range(self, lower: int | None, upper: int | None):
        if lower is None:
            lower = -MAX_VAL
        if upper is None:
            upper = MAX_VAL
        self.setRange(lower, upper)


if __name__ == "__main__":
    val = IntValidator()
    val.setRange(0, 9)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Literal

from prettyqt import core, pdf, widgets
from prettyqt.qt import QtPdfWidgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


PageModeStr = Literal[
    "single",
    "multi",
]

PAGE_MODE: bidict[PageModeStr, QtPdfWidgets.QPdfView.PageMode] = bidict(
    single=QtPdfWidgets.QPdfView.PageMode.SinglePage,
    multi=QtPdfWidgets.QPdfView.PageMode.MultiPage,
)

ZoomModeStr = Literal[
    "custom",
    "fit_to_width",
    "fit_in_view",
]

ZOOM_MODE: bidict[ZoomModeStr, QtPdfWidgets.QPdfView.ZoomMode] = bidict(
    custom=QtPdfWidgets.QPdfView.ZoomMode.Custom,
    fit_to_width=QtPdfWidgets.QPdfView.ZoomMode.FitToWidth,
    fit_in_view=QtPdfWidgets.QPdfView.ZoomMode.FitInView,
)


class PdfView(widgets.AbstractScrollAreaMixin, QtPdfWidgets.QPdfView):
    """PDF viewer widget ."""

    def __init__(self, parent: widgets.QWidget | None = None, **kwargs):
        super().__init__(parent, **kwargs)
        self.setDocument(pdf.PdfDocument(self))

    def get_document_margins(self) -> core.Margins:
        return core.Margins(self.documentMargins())

    def set_file(self, path: datatypes.PathType):
        doc = pdf.PdfDocument(self)
        doc.load(os.fspath(path))
        self.setDocument(doc)

    def set_page_mode(self, mode: PageModeStr | QtPdfWidgets.QPdfView.PageMode):
        """Set the page mode.

        Args:
            mode: page mode
        """
        self.setPageMode(PAGE_MODE.get_enum_value(mode))

    def get_page_mode(self) -> PageModeStr:
        """Return current page mode.

        Returns:
            page mode
        """
        return PAGE_MODE.inverse[self.pageMode()]

    def set_zoom_mode(self, mode: ZoomModeStr | QtPdfWidgets.QPdfView.ZoomMode):
        """Set the zoom mode.

        Args:
            mode: zoom mode
        """
        self.setZoomMode(ZOOM_MODE.get_enum_value(mode))

    def get_zoom_mode(self) -> ZoomModeStr:
        """Return current zoom mode.

        Returns:
            zoom mode
        """
        return ZOOM_MODE.inverse[self.zoomMode()]


if __name__ == "__main__":
    app = widgets.app()
    view = PdfView()
    view.show()
    app.exec()



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Literal

from prettyqt import core, widgets
from prettyqt.qt import QtQuickWidgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


RESIZE_MODES = bidict(
    view_to_root_object=QtQuickWidgets.QQuickWidget.ResizeMode.SizeViewToRootObject,
    root_object_to_view=QtQuickWidgets.QQuickWidget.ResizeMode.SizeRootObjectToView,
)

ResizeModeStr = Literal["view_to_root_object", "root_object_to_view"]

STATUS = bidict(
    null=QtQuickWidgets.QQuickWidget.Status.Null,
    ready=QtQuickWidgets.QQuickWidget.Status.Ready,
    loading=QtQuickWidgets.QQuickWidget.Status.Loading,
    error=QtQuickWidgets.QQuickWidget.Status.Error,
)

StatusStr = Literal["null", "ready", "loading", "error"]


class QuickWidget(widgets.WidgetMixin, QtQuickWidgets.QQuickWidget):
    """Class providing a widget for displaying a Qt Quick user interface."""

    def set_source(self, source: datatypes.UrlType | datatypes.PathType):
        if isinstance(source, os.PathLike):
            source = os.fspath(source)
        if isinstance(source, str):
            source = core.Url.fromLocalFile(source)
        self.setSource(source)

    def get_source(self) -> pathlib.Path:
        return pathlib.Path(self.source().toLocalFile())

    def get_status(self) -> StatusStr:
        return STATUS.inverse[self.status()]


if __name__ == "__main__":
    app = widgets.app()
    widget = QuickWidget()



from __future__ import annotations

import logging
from typing import Any

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class UnmeltProxyModel(core.AbstractProxyModel):
    """Proxy model to pivot a table from long format to wide format.

    Reverses the MeltProxyModel operation.

    === "Without proxy"

        ```py
        app = widgets.app()
        data = {
            "ID": ["A", "A", "B", "B"] * 2,
            "Variable": ["height", "weight"] * 4,
            "Value": [5.5, 130, 6.0, 150, 5.7, 140, 5.9, 145],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.show()
        ```

    === "With proxy"

        ```py
        app = widgets.app()
        data = {
            "ID": ["A", "A", "B", "B"] * 2,
            "Variable": ["height", "weight"] * 4,
            "Value": [5.5, 130, 6.0, 150, 5.7, 140, 5.9, 145],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.unmelt(id_columns=[0], variable_column=1, value_column=2)
        table.show()
        ```
    """

    ID = "unmelt"
    ICON = "mdi6.table-pivot"

    def __init__(
        self,
        id_columns: list[int],
        variable_column: int,
        value_column: int,
        **kwargs,
    ):
        """Initialize UnmeltProxyModel.

        Args:
            id_columns: List of column indices that identify unique rows
            variable_column: Column index containing variable names
            value_column: Column index containing values
            kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self._id_columns = id_columns
        self._variable_column = variable_column
        self._value_column = value_column
        self._unique_vars: list[str] = []
        self._unique_ids: list[tuple] = []

    def _update_cache(self):
        """Update internal cache of unique variables and IDs."""
        source = self.sourceModel()
        if source is None:
            return

        # Get unique variables
        variables = set()
        for row in range(source.rowCount()):
            idx = source.index(row, self._variable_column)
            var = str(source.data(idx, constants.DISPLAY_ROLE))
            variables.add(var)
        self._unique_vars = sorted(variables)

        # Get unique ID combinations
        id_combinations = set()
        for row in range(source.rowCount()):
            id_values = []
            for col in self._id_columns:
                idx = source.index(row, col)
                value = source.data(idx, constants.DISPLAY_ROLE)
                id_values.append(value)
            id_combinations.add(tuple(id_values))
        self._unique_ids = sorted(id_combinations)

    def sourceModelReset(self):
        """Handle source model reset."""
        self._update_cache()
        super().sourceModelReset()

    def setSourceModel(self, model):
        """Set source model and update cache."""
        result = super().setSourceModel(model)
        self._update_cache()
        return result

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of rows in the model."""
        return len(self._unique_ids)

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of columns in the model."""
        return len(self._id_columns) + len(self._unique_vars)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return header data for given section and orientation."""
        if orientation != constants.HORIZONTAL or role != constants.DISPLAY_ROLE:
            return super().headerData(section, orientation, role)

        if section < len(self._id_columns):
            # Return original header for ID columns
            return self.sourceModel().headerData(
                self._id_columns[section], constants.HORIZONTAL, role
            )
        # Return variable name for value columns
        var_idx = section - len(self._id_columns)
        return self._unique_vars[var_idx]

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return data for given index and role."""
        if not index.isValid():
            return None

        source = self.sourceModel()
        row, col = index.row(), index.column()

        # Handle ID columns
        if col < len(self._id_columns):
            id_values = self._unique_ids[row]
            return id_values[col]

        # Handle value columns
        var_name = self._unique_vars[col - len(self._id_columns)]
        id_values = self._unique_ids[row]

        # Find matching row in source model
        for source_row in range(source.rowCount()):
            # Check if ID values match
            matches = True
            for i, id_col in enumerate(self._id_columns):
                idx = source.index(source_row, id_col)
                if source.data(idx, constants.DISPLAY_ROLE) != id_values[i]:
                    matches = False
                    break

            # Check if variable matches
            if matches:
                var_idx = source.index(source_row, self._variable_column)
                if source.data(var_idx, constants.DISPLAY_ROLE) == var_name:
                    value_idx = source.index(source_row, self._value_column)
                    return source.data(value_idx, role)

        return None

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        """Map proxy index to source index."""
        if not proxy_index.isValid():
            return core.ModelIndex()

        row, col = proxy_index.row(), proxy_index.column()

        # ID columns map directly
        if col < len(self._id_columns):
            id_values = self._unique_ids[row]
            # Find matching row in source
            for source_row in range(self.sourceModel().rowCount()):
                matches = True
                for i, id_col in enumerate(self._id_columns):
                    idx = self.sourceModel().index(source_row, id_col)
                    if self.sourceModel().data(idx) != id_values[i]:
                        matches = False
                        break
                if matches:
                    return self.sourceModel().index(source_row, self._id_columns[col])

        return core.ModelIndex()

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        """Map source index to proxy index."""
        if not source_index.isValid():
            return core.ModelIndex()

        source_row, source_col = source_index.row(), source_index.column()

        # Handle ID columns
        if source_col in self._id_columns:
            # Get ID values for this row
            id_values = []
            for col in self._id_columns:
                idx = self.sourceModel().index(source_row, col)
                id_values.append(self.sourceModel().data(idx))
            id_tuple = tuple(id_values)

            try:
                proxy_row = self._unique_ids.index(id_tuple)
                proxy_col = self._id_columns.index(source_col)
                return self.createIndex(proxy_row, proxy_col)
            except ValueError:
                return core.ModelIndex()

        return core.ModelIndex()


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app(style="Windows")
    data = {
        "ID": ["A", "A", "B", "B"] * 2,
        "Variable": ["height", "weight"] * 4,
        "Value": [5.5, 130, 6.0, 150, 5.7, 140, 5.9, 145],
    }
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.setWindowTitle("UnmeltProxyModel example")
    table.set_icon("mdi6.table-pivot")
    table.show()
    table.resize(600, 130)
    table.h_header.resize_sections("stretch")
    table.proxifier.unmelt(id_columns=[0], variable_column=1, value_column=2)
    with app.debug_mode():
        app.exec()
